typedef struct Byte_Array
{
	byte* data;
	inta length;
} Byte_Array;
typedef struct Wave
{
	byte format;
	longo num_channels, sample_rate;
	inta size;
	void* data;
} Wave;
typedef struct Audio_Stream_Info
{
	byte format;
	longo num_channels, sample_rate;
	inta size;
} Audio_Stream_Info;
#define STBV_MALLOC audio_memory_allocate
#define DRWAV_MALLOC audio_memory_allocate
#define DRFLAC_MALLOC audio_memory_allocate
#define STBV_REALLOC audio_memory_reallocate
#define DRWAV_REALLOC audio_memory_reallocate
#define DRFLAC_REALLOC audio_memory_reallocate
#define DRWAV_FREE audio_memory_free
#define DRFLAC_FREE audio_memory_free
static byte* audio_memory;
static inta audio_memory_pos;
static inta audio_memory_size;
static void* audio_memory_allocate(inta size)
{
	inta m = audio_memory_pos % 8;
	if(m != 0)
		audio_memory_pos += (8 - m);

	if(audio_memory_pos+size > audio_memory_size)
	{
		print_error("audio memory allocation failed");
		return NULL;
	}
	void* r = &audio_memory[audio_memory_pos];
	audio_memory_pos += size;
	return r;
}
static void* audio_memory_reallocate(void* data, inta old_size, inta size)
{
	if(((byte*)data + old_size) == ((byte*)audio_memory + audio_memory_pos))
	{
		audio_memory_pos += (size - old_size);

		if(audio_memory_pos >= audio_memory_size)
		{
			print_error("audio memory reallocation failed");
			return NULL;
		}
		return data;
	}
	void* new_data = audio_memory_allocate(size);
	memcpy(new_data, data, old_size);
	return new_data;
}
static void audio_memory_free(void* data) {}
#define STB_VORBIS_NO_PUSHDATA_API
#define DR_WAV_IMPLEMENTATION
#define DR_FLAC_IMPLEMENTATION
#include "dr_wav.h"
#include "dr_flac.h"
#include "stb_vorbis.h"
void audio_wave_load(octa* r, byte format, octa* file_name_buf)
{
	String file_name;
	memcpy(&file_name, file_name_buf, 16);
	if(string_equals(string_find_extension(file_name), make_string_c("wav")))
	{
		Wave a;
		unsigned int channels;
		unsigned int sample_rate;
		drwav_uint64 num_frames;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			a.data = drwav_open_file_and_read_pcm_frames_s16(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 2;
			a.format = AUDIO_FORMAT_SHORT;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		{
			a.data = drwav_open_file_and_read_pcm_frames_s32(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_LONG;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_FLOAT:
		{
			a.data = drwav_open_file_and_read_pcm_frames_f32(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_FLOAT;
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		a.num_channels = channels;
		a.sample_rate = sample_rate;
		memcpy(r, &a, sizeof(Wave));
	}
	else if(string_equals(string_find_extension(file_name), make_string_c("flac")))
	{
		Wave a;
		unsigned int channels;
		unsigned int sample_rate;
		drflac_uint64 num_frames;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			a.data = drflac_open_file_and_read_pcm_frames_s16(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 2;
			a.format = AUDIO_FORMAT_SHORT;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		{
			a.data = drflac_open_file_and_read_pcm_frames_s32(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_LONG;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_FLOAT:
		{
			a.data = drflac_open_file_and_read_pcm_frames_f32(
				to_cstr(file_name), &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_FLOAT;
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		a.num_channels = channels;
		a.sample_rate = sample_rate;
		memcpy(r, &a, sizeof(Wave));
	}
	else if(string_equals(string_find_extension(file_name), make_string_c("ogg")))
	{
		Wave a;
		int channels = 0;
		int sample_rate = 0;
		inta num_frames = 0;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			short* samples = NULL;
			num_frames = stb_vorbis_decode_filename(
				to_cstr(file_name), &channels, &sample_rate, &samples);
			if(num_frames >= 0)
			{
				a.format = AUDIO_FORMAT_SHORT;
				a.num_channels = channels;
				a.sample_rate = sample_rate;
				a.data = samples;
				a.size = num_frames * channels * 2;
				audio_wave_convert((octa*)&a, format);
			}
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		case AUDIO_FORMAT_FLOAT:
		{
			int err = 0;
			stb_vorbis* ogg = stb_vorbis_open_filename(to_cstr(file_name), &err, NULL);
			if(ogg != NULL)
			{
				stb_vorbis_info vorbis = stb_vorbis_get_info(ogg);
				num_frames = stb_vorbis_stream_length_in_samples(ogg);
				a.format = AUDIO_FORMAT_FLOAT;
				a.num_channels = vorbis.channels;
				a.sample_rate = vorbis.sample_rate;
				a.size = num_frames * vorbis.channels * 4;
				a.data = audio_memory_allocate(a.size);
				int numframes2 = stb_vorbis_get_samples_float_interleaved(
					ogg, vorbis.channels, a.data, num_frames * vorbis.channels);
				audio_wave_convert((octa*)&a, format);
				stb_vorbis_close(ogg);
			}
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		memcpy(r, &a, sizeof(Wave));
	}
	else
		memset(r, 0, sizeof(Wave));
}
void audio_wave_load_from_memory(octa* r, byte format, inta size, octa data)
{
	byte* bytes = (byte*)data;
	inta length = size;
	if(length < 4)
	{
		memset(r, 0, sizeof(Wave));
		return;
	}
	if(bytes[0] == 'R' && bytes[1] == 'I' && bytes[2] == 'F' && bytes[3] == 'F')
	{
		Wave a;
		unsigned int channels;
		unsigned int sample_rate;
		drwav_uint64 num_frames;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			a.data = drwav_open_memory_and_read_pcm_frames_s16(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 2;
			a.format = AUDIO_FORMAT_SHORT;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		{
			a.data = drwav_open_memory_and_read_pcm_frames_s32(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_LONG;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_FLOAT:
		{
			a.data = drwav_open_memory_and_read_pcm_frames_f32(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_FLOAT;
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		a.num_channels = channels;
		a.sample_rate = sample_rate;
		memcpy(r, &a, sizeof(Wave));
	}
	else if(bytes[0] == 'f' && bytes[1] == 'L' && bytes[2] == 'a' && bytes[3] == 'C')
	{
		Wave a;
		unsigned int channels;
		unsigned int sample_rate;
		drwav_uint64 num_frames;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			a.data = drflac_open_memory_and_read_pcm_frames_s16(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 2;
			a.format = AUDIO_FORMAT_SHORT;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		{
			a.data = drflac_open_memory_and_read_pcm_frames_s32(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_LONG;
			audio_wave_convert((octa*)&a, format);
			break;
		}
		case AUDIO_FORMAT_FLOAT:
		{
			a.data = drflac_open_memory_and_read_pcm_frames_f32(
				bytes, length, &channels, &sample_rate, &num_frames);
			a.size = num_frames * channels * 4;
			a.format = AUDIO_FORMAT_FLOAT;
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		a.num_channels = channels;
		a.sample_rate = sample_rate;
		memcpy(r, &a, sizeof(Wave));
	}
	else if(bytes[0] == 'O' && bytes[1] == 'g' && bytes[2] == 'g' && bytes[3] == 'S')
	{
		Wave a;
		int channels = 0;
		int sample_rate = 0;
		int num_frames;
		switch(format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
		{
			short* samples = NULL;
			num_frames = stb_vorbis_decode_memory(
				bytes, length, &channels, &sample_rate, &samples);
			if(num_frames >= 0)
			{
				a.format = AUDIO_FORMAT_SHORT;
				a.num_channels = channels;
				a.sample_rate = sample_rate;
				a.data = samples;
				a.size = num_frames * channels * 2;
				audio_wave_convert((octa*)&a, format);
			}
			break;
		}
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		case AUDIO_FORMAT_FLOAT:
		{
			int err = 0;
			stb_vorbis* ogg = stb_vorbis_open_memory(bytes, length, &err, NULL);
			if(ogg != NULL)
			{
				stb_vorbis_info vorbis = stb_vorbis_get_info(ogg);
				num_frames = stb_vorbis_stream_length_in_samples(ogg);
				a.format = AUDIO_FORMAT_FLOAT;
				a.num_channels = vorbis.channels;
				a.sample_rate = vorbis.sample_rate;
				a.size = num_frames * vorbis.channels * 4;
				a.data = audio_memory_allocate(a.size);
				stb_vorbis_get_samples_float_interleaved(
					ogg, vorbis.channels, a.data, num_frames * vorbis.channels);
				audio_wave_convert((octa*)&a, format);
				stb_vorbis_close(ogg);
			}
			break;
		}
		default:
			memset(r, 0, sizeof(Wave));
			return;
		}
		memcpy(r, &a, sizeof(Wave));
	}
	else
		memset(r, 0, sizeof(Wave));
}
static int calculate_sample_size(byte format)
{
	switch(format)
	{
	case AUDIO_FORMAT_BYTE:
	case AUDIO_FORMAT_TINY:
		return 1;
	case AUDIO_FORMAT_QUADRO:
	case AUDIO_FORMAT_FLOAT:
	case AUDIO_FORMAT_LONG:
		return 4;
	}
	return 2;
}
static void convert_wave_to_signed_integers(Wave* a)
{
	switch(a->format)
	{
	case AUDIO_FORMAT_BYTE:
		audio_wave_convert((octa*)a, AUDIO_FORMAT_TINY);
		break;
	case AUDIO_FORMAT_DOUBLE:
		audio_wave_convert((octa*)a, AUDIO_FORMAT_SHORT);
		break;
	case AUDIO_FORMAT_QUADRO:
	case AUDIO_FORMAT_FLOAT:
		audio_wave_convert((octa*)a, AUDIO_FORMAT_LONG);
		break;
	}
}
byte audio_wave_save(octa* a_buf, octa* file_name_buf)
{
	Wave a;
	String file_name;
	memcpy(&a, a_buf, sizeof(Wave));
	memcpy(&file_name, file_name_buf, 16);
	drwav wav;
	drwav_data_format format;
	//we need 64 bits if num_samples does not fit in 32 bits
	if(a.size < 0x100000000)
		format.container = drwav_container_riff;
	else
		format.container = drwav_container_w64;
	format.format = DR_WAVE_FORMAT_PCM; //basic format
	format.channels = a.num_channels;
	format.sampleRate = a.sample_rate;
	int sample_size = calculate_sample_size(a.format);
	convert_wave_to_signed_integers(&a);
	format.bitsPerSample = sample_size * 8;
	inta num_samples = a.size / sample_size;
	if(!drwav_init_file_write_sequential(&wav, to_cstr(file_name), &format, num_samples))
		return false;
	inta num = drwav_write_pcm_frames(&wav, num_samples / a.num_channels, a.data);
	drwav_uninit(&wav);
	return num != 0;
}
typedef union Audio_Stream_Ptr
{
	drwav* wav;
	drflac* flac;
	stb_vorbis* ogg;
} Audio_Stream_Ptr;
typedef union Audio_Stream_Source
{
	String file_name;
	Byte_Array data;
} Audio_Stream_Source;
typedef struct Internal_Audio_Stream
{
	byte alive;
	byte kind;
	byte writing, is_file;
	byte format, has_info;
	Audio_Stream_Ptr stream;
	Audio_Stream_Source source;
	Audio_Stream_Info info;
} Internal_Audio_Stream;
#define MAX_AUDIO_STREAMS 64
Internal_Audio_Stream audio_streams[MAX_AUDIO_STREAMS];
longo num_audio_streams;
void audio_set_buffer(inta size, octa buffer)
{
	audio_memory = (byte*)buffer;
	audio_memory_size = size;
	audio_memory_pos = 0;
}
void audio_set_buffer_position(inta pos)
{
	audio_memory_pos = pos;
}
inta audio_get_buffer_position()
{
	return audio_memory_pos;
}
longo audio_stream_open(byte format, octa* file_name_buf, byte open_for_writing)
{
	if(format == AUDIO_FORMAT_BYTE || format == AUDIO_FORMAT_TINY)
		return -1;
	String file_name;
	memcpy(&file_name, file_name_buf, 16);
	longo h = 0;
	for(; h < num_audio_streams; h++)
		if(!audio_streams[h].alive)
			break;
	if(h >= MAX_AUDIO_STREAMS)
		return -1;
	if(h >= num_audio_streams)
		num_audio_streams = h+1;
	Internal_Audio_Stream* stream = &audio_streams[h];
	stream->alive = true;
	stream->is_file = true;
	stream->source.file_name = file_name;
	stream->kind = AUDIO_STREAM_WAV;
	stream->stream.wav = NULL;
	stream->writing = open_for_writing;
	stream->format = format;
	stream->has_info = false;
	if(!open_for_writing)
	{
		if(string_equals(string_find_extension(file_name), make_string_c("ogg")))
		{
			stream->kind = AUDIO_STREAM_OGG;
			stream->stream.ogg = NULL;
		}
		else if(string_equals(string_find_extension(file_name), make_string_c("flac")))
		{
			stream->kind = AUDIO_STREAM_FLAC;
			stream->stream.flac = NULL;
		}
	}
	return h;
}
longo audio_stream_open_from_memory(byte format, inta size, octa data)
{
	if(format == AUDIO_FORMAT_BYTE || format == AUDIO_FORMAT_TINY)
		return -1;
	byte* bytes = (byte*)data;
	inta length = size;
	longo h = 0;
	for(; h < num_audio_streams; h++)
		if(!audio_streams[h].alive)
			break;
	if(h >= MAX_AUDIO_STREAMS)
		return -1;
	if(h >= num_audio_streams)
		num_audio_streams = h+1;
	Internal_Audio_Stream* stream = &audio_streams[h];
	stream->alive = true;
	stream->is_file = false;
	stream->source.data.data = bytes;
	stream->source.data.length = length;
	stream->kind = AUDIO_STREAM_WAV;
	stream->stream.wav = NULL;
	stream->writing = false;
	stream->format = format;
	stream->has_info = false;
	if(length >= 4)
	{
		if(bytes[0] == 'O' && bytes[1] == 'g' && bytes[2] == 'g' && bytes[3] == 'S')
		{
			stream->kind = AUDIO_STREAM_OGG;
			stream->stream.ogg = NULL;
		}
		else if(bytes[0] == 'f' && bytes[1] == 'L' && bytes[2] == 'a' && bytes[3] == 'C')
		{
			stream->kind = AUDIO_STREAM_FLAC;
			stream->stream.flac = NULL;
		}
	}
	return h;
}
void audio_stream_close(longo a)
{
	switch(audio_streams[a].kind)
	{
	case AUDIO_STREAM_WAV:
		if(audio_streams[a].stream.wav != NULL)
			drwav_close(audio_streams[a].stream.wav);
		break;
	case AUDIO_STREAM_FLAC:
		if(audio_streams[a].stream.flac != NULL)
			drflac_close(audio_streams[a].stream.flac);
		break;
	case AUDIO_STREAM_OGG:
		if(audio_streams[a].stream.ogg != NULL)
			stb_vorbis_close(audio_streams[a].stream.ogg);
		break;
	}
	audio_streams[a].alive = false;
}
void audio_stream_read_info(octa* r, longo a)
{
	if(audio_streams[a].has_info)
	{
		memcpy(r, &audio_streams[a].info, sizeof(Audio_Stream_Info));
		return;
	}
	Audio_Stream_Info info;
	memset(&info, 0, sizeof(Audio_Stream_Info));
	info.format = audio_streams[a].format;
	int sample_size = calculate_sample_size(info.format);
	switch(audio_streams[a].kind)
	{
	case AUDIO_STREAM_WAV:
		if(audio_streams[a].is_file)
			audio_streams[a].stream.wav = drwav_open_file(
				to_cstr(audio_streams[a].source.file_name));
		else
			audio_streams[a].stream.wav = drwav_open_memory(
				audio_streams[a].source.data.data,
				audio_streams[a].source.data.length);
		if(audio_streams[a].stream.wav != NULL)
		{
			drwav* wav = audio_streams[a].stream.wav;
			info.num_channels = wav->channels;
			info.sample_rate = wav->sampleRate;
			info.size = wav->totalPCMFrameCount * info.num_channels * sample_size;
		}
		break;
	case AUDIO_STREAM_FLAC:
		if(audio_streams[a].is_file)
			audio_streams[a].stream.flac = drflac_open_file(
				to_cstr(audio_streams[a].source.file_name));
		else
			audio_streams[a].stream.flac = drflac_open_memory(
				audio_streams[a].source.data.data,
				audio_streams[a].source.data.length);
		if(audio_streams[a].stream.flac != NULL)
		{
			drflac* flac = audio_streams[a].stream.flac;
			info.num_channels = flac->channels;
			info.sample_rate = flac->sampleRate;
			info.size = flac->totalPCMFrameCount * info.num_channels * sample_size;
		}
		break;
	case AUDIO_STREAM_OGG:
	{
		int err = 0;
		if(audio_streams[a].is_file)
			audio_streams[a].stream.ogg = stb_vorbis_open_filename(
				to_cstr(audio_streams[a].source.file_name), &err, NULL);
		else
			audio_streams[a].stream.ogg = stb_vorbis_open_memory(
				audio_streams[a].source.data.data,
				audio_streams[a].source.data.length, &err, NULL);
		if(audio_streams[a].stream.ogg != NULL)
		{
			stb_vorbis_info vorbis = stb_vorbis_get_info(audio_streams[a].stream.ogg);
			info.size = stb_vorbis_stream_length_in_samples(audio_streams[a].stream.ogg)
				* vorbis.channels * sample_size;
			info.num_channels = vorbis.channels;
			info.sample_rate = vorbis.sample_rate;
		}
		break;
	}
	}
	audio_streams[a].has_info = true;
	audio_streams[a].info = info;
	memcpy(r, &info, sizeof(Audio_Stream_Info));
}
void audio_stream_write_info(longo a, octa* info_buf)
{
	if(audio_streams[a].has_info)
		return;
	Audio_Stream_Info info = *(Audio_Stream_Info*)info_buf;
	drwav_data_format format;
	if(info.size < 0x100000000)
		format.container = drwav_container_riff;
	else
		format.container = drwav_container_w64;
	format.format = DR_WAVE_FORMAT_PCM;
	format.channels = info.num_channels;
	format.sampleRate = info.sample_rate;
	switch(info.format)
	{
	case AUDIO_FORMAT_BYTE: info.format = AUDIO_FORMAT_TINY; break;
	case AUDIO_FORMAT_DOUBLE: info.format = AUDIO_FORMAT_SHORT; break;
	case AUDIO_FORMAT_QUADRO:
	case AUDIO_FORMAT_FLOAT: info.format = AUDIO_FORMAT_LONG; break;
	}
	int sample_size = calculate_sample_size(info.format);
	format.bitsPerSample = sample_size * 8;
	inta num_samples = info.size / sample_size;
	if(num_samples > 0)
		audio_streams[a].stream.wav = drwav_open_file_write_sequential(
			to_cstr(audio_streams[a].source.file_name), &format, num_samples);
	else
		audio_streams[a].stream.wav = drwav_open_file_write(
			to_cstr(audio_streams[a].source.file_name), &format);
	audio_streams[a].has_info = true;
	audio_streams[a].info = info;
}
inta audio_stream_write(longo a, inta num_frames, octa data)
{
	Wave wave;
	wave.format = audio_streams[a].format;
	wave.num_channels = audio_streams[a].info.num_channels;
	wave.sample_rate = audio_streams[a].info.sample_rate;
	wave.size = calculate_sample_size(wave.format);
	wave.data = (void*)data;
	audio_wave_convert((octa*)&wave, audio_streams[a].info.format);
	return drwav_write_pcm_frames(audio_streams[a].stream.wav, num_frames, (void*)wave.data);
}
inta audio_stream_read(longo a, inta num_frames, octa data)
{
	Wave wave;
	wave.num_channels = audio_streams[a].info.num_channels;
	wave.sample_rate = audio_streams[a].info.sample_rate;
	wave.data = (void*)data;
	inta num_frames_read = 0;
	switch(audio_streams[a].kind)
	{
	case AUDIO_STREAM_WAV:
	{
		switch(audio_streams[a].format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
			wave.format = AUDIO_FORMAT_SHORT;
			wave.size = num_frames * wave.num_channels * 2;
			num_frames_read = drwav_read_pcm_frames_s16(
				audio_streams[a].stream.wav, num_frames, (shorti*)wave.data);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
			wave.format = AUDIO_FORMAT_LONG;
			wave.size = num_frames * wave.num_channels * 4;
			num_frames_read = drwav_read_pcm_frames_s32(
				audio_streams[a].stream.wav, num_frames, (longo*)wave.data);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		case AUDIO_FORMAT_FLOAT:
			wave.format = AUDIO_FORMAT_FLOAT;
			wave.size = num_frames * wave.num_channels * 4;
			num_frames_read = drwav_read_pcm_frames_f32(
				audio_streams[a].stream.wav, num_frames, (float*)wave.data);
			break;
		default:
			print_error("invalid audio format");
		}
		break;
	}
	case AUDIO_STREAM_FLAC:
	{
		switch(audio_streams[a].format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
			wave.format = AUDIO_FORMAT_SHORT;
			wave.size = num_frames * wave.num_channels * 2;
			num_frames_read = drflac_read_pcm_frames_s16(
				audio_streams[a].stream.flac, num_frames, (shorti*)wave.data);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
			wave.format = AUDIO_FORMAT_LONG;
			wave.size = num_frames * wave.num_channels * 4;
			num_frames_read = drflac_read_pcm_frames_s32(
				audio_streams[a].stream.flac, num_frames, (longo*)wave.data);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		case AUDIO_FORMAT_FLOAT:
			wave.format = AUDIO_FORMAT_FLOAT;
			wave.size = num_frames * wave.num_channels * 4;
			num_frames_read = drflac_read_pcm_frames_f32(
				audio_streams[a].stream.flac, num_frames, (float*)wave.data);
			break;
		default:
			print_error("invalid audio format");
		}
		break;
	}
	case AUDIO_STREAM_OGG:
	{
		inta num_samples = num_frames * audio_streams[a].info.num_channels;
		switch(audio_streams[a].format)
		{
		case AUDIO_FORMAT_BYTE:
		case AUDIO_FORMAT_TINY:
		case AUDIO_FORMAT_DOUBLE:
		case AUDIO_FORMAT_SHORT:
			wave.format = AUDIO_FORMAT_SHORT;
			wave.size = num_frames * wave.num_channels * 2;
			num_frames_read = stb_vorbis_get_samples_short_interleaved(
				audio_streams[a].stream.ogg, wave.num_channels, (void*)wave.data, num_samples);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		case AUDIO_FORMAT_QUADRO:
		case AUDIO_FORMAT_LONG:
		case AUDIO_FORMAT_FLOAT:
			wave.format = AUDIO_FORMAT_FLOAT;
			wave.size = num_frames * wave.num_channels * 4;
			num_frames_read = stb_vorbis_get_samples_float_interleaved(
				audio_streams[a].stream.ogg, wave.num_channels, (void*)wave.data, num_samples);
			audio_wave_convert((octa*)&wave, audio_streams[a].format);
			break;
		default:
			print_error("invalid audio format");
		}
	}
	}
	return num_frames_read;
}
byte audio_stream_seek(longo a, inta frame)
{
	switch(audio_streams[a].kind)
	{
	case AUDIO_STREAM_WAV:
		return drwav_seek_to_pcm_frame(
			audio_streams[a].stream.wav, frame);
	case AUDIO_STREAM_FLAC:
		return drflac_seek_to_pcm_frame(
			audio_streams[a].stream.flac, frame);
	case AUDIO_STREAM_OGG:
		return stb_vorbis_seek(
			audio_streams[a].stream.ogg, audio_streams[a].info.num_channels * frame);
	}
	return false;
}
void audio_wave_convert(octa* a, byte format)
{
	Wave* wave = (Wave*)a;
	if(wave->format == format) return;
	int from_size = calculate_sample_size(wave->format);
	int from_is_float = wave->format == AUDIO_FORMAT_FLOAT;
	int from_is_signed = (wave->format == AUDIO_FORMAT_TINY || wave->format == AUDIO_FORMAT_SHORT || wave->format == AUDIO_FORMAT_LONG);
	int to_size = calculate_sample_size(format);
	int to_is_float = format == AUDIO_FORMAT_FLOAT;
	int to_is_signed = (format == AUDIO_FORMAT_TINY || format == AUDIO_FORMAT_SHORT || format == AUDIO_FORMAT_LONG);
	if(!from_is_float && !to_is_float)
	{
		//integer->integer
		if(from_size != to_size)
		{
			//big->small
			if(from_size == 2 && to_size == 1)
			{
				byte* bytes = (byte*)wave->data;
				doubli* doubles = (doubli*)wave->data;
				inta num_samples = wave->size / 2;
				for(inta i = 0; i < num_samples; i++)
					bytes[i] = doubles[i] >> 8;
			}
			else if(from_size == 4 && to_size == 1)
			{
				byte* bytes = (byte*)wave->data;
				quadro* quadros = (quadro*)wave->data;
				inta num_samples = wave->size / 4;
				for(inta i = 0; i < num_samples; i++)
					bytes[i] = quadros[i] >> 24;
			}
			else if(from_size == 4 && to_size == 2)
			{
				doubli* doubles = (doubli*)wave->data;
				quadro* quadros = (quadro*)wave->data;
				inta num_samples = wave->size / 4;
				for(inta i = 0; i < num_samples; i++)
					doubles[i] = quadros[i] >> 16;
			}
			//small->big
			else if(from_size == 1 && to_size == 2)
			{
				byte* bytes = (byte*)wave->data;
				doubli* doubles = (doubli*)wave->data;
				inta num_samples = wave->size;
				for(inta i = num_samples-1; i >= 0; i--)
					doubles[i] = (shorti)((tiny)bytes[i]) << 8;
			}
			else if(from_size == 1 && to_size == 4)
			{
				byte* bytes = (byte*)wave->data;
				quadro* quadros = (quadro*)wave->data;
				inta num_samples = wave->size;
				for(inta i = num_samples-1; i >= 0; i--)
					quadros[i] = (longo)((tiny)bytes[i]) << 24;
			}
			else if(from_size == 2 && to_size == 4)
			{
				doubli* doubles = (doubli*)wave->data;
				quadro* quadros = (quadro*)wave->data;
				inta num_samples = wave->size / 2;
				for(inta i = num_samples-1; i >= 0; i--)
					quadros[i] = (longo)((shorti)doubles[i]) << 16;
			}
		}
		//unsigned<->signed
		if(from_is_signed != to_is_signed)
		{
			byte* bytes = (byte*)wave->data;
			for(inta i = to_size-1; i < wave->size; i += to_size)
				bytes[i] ^= 0x80;
		}
	}
	else if(from_is_float && !to_is_float)
	{
		//float->integer
		float* floats = (float*)wave->data;
		inta num_samples = wave->size / 4;
		if(to_size == 4)
		{
			quadro* samples = (quadro*)wave->data;
			for(inta i = 0; i < num_samples; i++)
				samples[i] = (int)((double)floats[i] * (double)0x7fffffff);
		}
		else if(to_size == 2)
		{
			doubli* samples = (doubli*)wave->data;
			for(inta i = 0; i < num_samples; i++)
				samples[i] = (int)((double)floats[i] * (double)0x7fff);
		}
		else //to_size == 1
		{
			byte* samples = (byte*)wave->data;
			for(inta i = 0; i < num_samples; i++)
				samples[i] = (int)((double)floats[i] * (double)0x7f);
		}
		if(!to_is_signed)
		{
			byte* bytes = (byte*)wave->data;
			for(inta i = to_size-1; i < wave->size; i += to_size)
				bytes[i] ^= 0x80;
		}
	}
	else if(!from_is_float && to_is_float)
	{
		//integer->float
		if(!from_is_signed)
		{
			byte* bytes = (byte*)wave->data;
			for(inta i = from_size-1; i < wave->size; i += from_size)
				bytes[i] ^= 0x80;
		}
		float* floats = (float*)wave->data;
		if(from_size == 4)
		{
			inta num_samples = wave->size / 4;
			longo* samples = (longo*)wave->data;
			for(inta i = num_samples-1; i >= 0; i--)
				floats[i] = (float)((double)(int)samples[i] / (double)0x7fffffff);
		}
		else if(from_size == 2)
		{
			inta num_samples = wave->size / 2;
			shorti* samples = (shorti*)wave->data;
			for(inta i = num_samples-1; i >= 0; i--)
				floats[i] = (float)((double)(int)samples[i] / (double)0x7fff);
		}
		else //from_size == 1
		{
			inta num_samples = wave->size;
			tiny* samples = (tiny*)wave->data;
			for(inta i = num_samples-1; i >= 0; i--)
				floats[i] = (float)((double)(int)samples[i] / (double)0x7f);
		}
	}
	wave->format = format;
	if(to_size > from_size)
		wave->size *= (to_size / from_size);
	else if(to_size < from_size)
		wave->size /= (from_size / to_size);
}
