//enums
#ifdef WINDOWS_OS
wchar_t* to_wide_cstr(String s)
{
	int len = MultiByteToWideChar(CP_UTF8, 0, s.data, s.length, NULL, 0);
	int num_bytes = 2*(len+1);
	if(temporary_pos + num_bytes > MAX_TEMP_SIZE)
		temporary_pos = 0;
	char* m = &temporary_buffer[temporary_pos];
	MultiByteToWideChar(CP_UTF8, 0, s.data, s.length, (wchar_t*)m, len);
	m[num_bytes-2] = 0;
	m[num_bytes-1] = 0;
	temporary_pos += num_bytes;
	return (wchar_t*)m;
}
byte file_copy(octa* src, octa* dest)
{
	String a, b;
	memcpy(&a, src, 16);
	memcpy(&b, dest, 16);
	FILE* file = _wfopen(to_wide_cstr(a), to_wide_cstr(make_string_c("r")));
	if(file == NULL)
		return false;
	FILE* target = _wfopen(to_wide_cstr(b), to_wide_cstr(make_string_c("w")));
	if(target == NULL)
	{
		fclose(file);
		return false;
	}
	int c;
	while((c = fgetc(file)) != EOF)
		fputc(c, target);
	fclose(file);
	fclose(target);
	return true;
}
byte file_create(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	FILE* file = _wfopen(to_wide_cstr(s), to_wide_cstr(make_string_c("wb")));
	if(file != NULL)
	{
		fclose(file);
		return true;
	}
	else
		return false;
}
octa file_open(octa* name, byte flags)
{
	String s;
	memcpy(&s, name, 16);
	const wchar_t* cname = to_wide_cstr(s);
	FILE* fp = NULL;
	if(flags == FILE_WRITE)
		fp = _wfopen(cname, to_wide_cstr(make_string_c("wb")));
	else if(flags == FILE_READ)
		fp = _wfopen(cname, to_wide_cstr(make_string_c("rb")));
	else if(flags == (FILE_READ | FILE_WRITE))
		fp = _wfopen(cname, to_wide_cstr(make_string_c("r+b")));
	else if(flags == FILE_APPEND)
		fp = _wfopen(cname, to_wide_cstr(make_string_c("a")));
	else if(flags == (FILE_READ | FILE_APPEND))
		fp = _wfopen(cname, to_wide_cstr(make_string_c("a+")));
	return (octa)fp;
}
#endif
#ifdef LINUX_OS
byte file_copy(octa* src, octa* dest)
{
	String a, b;
	memcpy(&a, src, 16);
	memcpy(&b, dest, 16);
	FILE* file = fopen(to_cstr(a), "r");
	if(file == NULL)
		return false;
	FILE* target = fopen(to_cstr(b), "w");
	if(target == NULL)
	{
		fclose(file);
		return false;
	}
	int c;
	while((c = fgetc(file)) != EOF)
		fputc(c, target);
	fclose(file);
	fclose(target);
	return true;
}
byte file_create(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	FILE* file = fopen(to_cstr(s), "wb");
	if(file != NULL)
	{
		fclose(file);
		return true;
	}
	else
		return false;
}
octa file_open(octa* name, byte flags)
{
	String s;
	memcpy(&s, name, 16);
	const char* cname = to_cstr(s);
	FILE* fp = NULL;
	if(flags == FILE_WRITE)
		fp = fopen(cname, "wb");
	else if(flags == FILE_READ)
		fp = fopen(cname, "rb");
	else if(flags == (FILE_READ | FILE_WRITE))
		fp = fopen(cname, "r+b");
	else if(flags == FILE_APPEND)
		fp = fopen(cname, "a");
	else if(flags == (FILE_READ | FILE_APPEND))
		fp = fopen(cname, "a+");
	return (octa)fp;
}
#endif
void file_close(octa handle)
{
	fclose((FILE*)handle);
}
byte file_failed(octa handle)
{
	return handle == 0;
}
byte file_seek(octa handle, inta offset)
{
	return fseeko((FILE*)handle, offset, SEEK_SET) == 0;
}
byte file_end(octa handle)
{
	return feof((FILE*)handle);
}
inta file_size(octa handle)
{
	FILE* file = (FILE*)handle;
	inta prev = ftello(file);
	fseeko(file, 0, SEEK_END);
	inta size = ftello(file);
	fseeko(file, prev, SEEK_SET);
	return size;
}
byte file_exists(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	FILE* file = fopen(to_cstr(s), "r");
	if(file != NULL)
	{
		fclose(file);
		return true;
	}
	else
		return false;
}
inta file_write(octa handle, inta size, octa data)
{
	return fwrite((byte*)data, 1, size, (FILE*)handle);
}
void file_write_byte(octa handle, byte b)
{
	fputc(b, (FILE*)handle);
}
inta file_read(octa handle, inta size, octa data)
{
	return fread((byte*)data, 1, size, (FILE*)handle);
}
byte file_read_byte(octa handle)
{
	return fgetc((FILE*)handle);
}
static void add_string(String** strings, inta* length, inta* capacity, String s)
{
	if((*capacity) == 0)
	{
		*capacity = 8;
		*strings = (String*)ALLOC((*capacity) * sizeof(String));
	}
	if((*length) >= (*capacity))
	{
		inta old_cap = *capacity;
		(*capacity) *= 2;
		*strings = (String*)REALLOC(*strings,
			(*capacity) * sizeof(String), old_cap * sizeof(String));
	}
	(*strings)[*length] = s;
	(*length)++;
}
typedef struct Directory_Info
{
	String* directory_names;
	inta num_directory_names;
	String* file_names;
	inta num_file_names;
} Directory_Info;
#ifdef LINUX_OS
static int is_directory(const char* path)
{
	struct stat statbuf;
	if(stat(path, &statbuf) != 0)
		return 0;
	return S_ISDIR(statbuf.st_mode);
}
byte file_delete(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	const char* cname = to_cstr(s);
	if(is_directory(cname))
		return false;
	return remove(cname) == 0;
}
byte file_rename(octa* name, octa* new_name)
{
	String a, b;
	memcpy(&a, name, 16);
	memcpy(&b, new_name, 16);
	const char* cname = to_cstr(a);
	if(is_directory(cname))
		return false;
	return rename(cname, to_cstr(b)) == 0;
}
byte directory_create(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return mkdir(to_cstr(s), 0777) == 0;
}
byte directory_delete(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return rmdir(to_cstr(s)) == 0;
}
byte directory_rename(octa* name, octa* new_name)
{
	String a, b;
	memcpy(&a, name, 16);
	memcpy(&b, new_name, 16);
	const char* c_old_name = to_cstr(a);
	if(!is_directory(c_old_name))
		return false;
	return rename(c_old_name, to_cstr(b)) == 0;
}
byte directory_exists(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return is_directory(to_cstr(s));
}
void directory_get_content(octa* r, octa* name)
{
	String s;
	memcpy(&s, name, 16);
	Directory_Info info;
	memset(&info, 0, sizeof(Directory_Info));
	inta dcap = 0, fcap = 0;
	DIR* dp;
	struct dirent* ep;
	if(s.length == 0)
		dp = opendir("./");
	else
		dp = opendir(to_cstr(s));
	if(dp != NULL)
	{
		while((ep = readdir(dp)))
		{
			const char* local_name = ep->d_name;
			int len = strlen(local_name);
			char* dest = (char*)ALLOC(len);
			memcpy(dest, local_name, len);
			if(ep->d_type == DT_DIR)
			{
				add_string(&info.directory_names, &info.num_directory_names, &dcap, make_string(dest, len));
			}
			else
			{
				add_string(&info.file_names, 
					&info.num_file_names, &fcap, make_string(dest, len));
			}
		}
		closedir(dp);
	}
	memcpy(r, &info, sizeof(Directory_Info));
}
#endif
#ifdef WINDOWS_OS
static byte is_directory(const wchar_t* path)
{
  DWORD dwAttrib = GetFileAttributesW(path);
  return (dwAttrib != INVALID_FILE_ATTRIBUTES
		&& (dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}
byte file_delete(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return DeleteFileW(to_wide_cstr(s));
}
byte file_rename(octa* old_name, octa* new_name)
{
	String a, b;
	memcpy(&a, old_name, 16);
	memcpy(&b, new_name, 16);
	const wchar_t* src = to_wide_cstr(a);
	const wchar_t* dest = to_wide_cstr(b);
	if(is_directory(src))
		return false;
	return MoveFileW(src, dest);
}
byte directory_create(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return CreateDirectoryW(to_wide_cstr(s), NULL);
}
byte directory_delete(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return RemoveDirectoryW(to_wide_cstr(s));
}
byte directory_rename(octa* old_name, octa* new_name)
{
	String a, b;
	memcpy(&a, old_name, 16);
	memcpy(&b, new_name, 16);
	const wchar_t* src = to_wide_cstr(a);
	const wchar_t* dest = to_wide_cstr(b);
	if(is_directory(src))
		return MoveFileW(src, dest);
	return false;
}
byte directory_exists(octa* name)
{
	String s;
	memcpy(&s, name, 16);
	return is_directory(to_wide_cstr(s));
}
void directory_get_content(octa* r, octa* name)
{
	String s;
	memcpy(&s, name, 16);
	Directory_Info info;
	memset(&info, 0, sizeof(Directory_Info));
	inta dcap = 0, fcap = 0;
	WIN32_FIND_DATAW data;
	HANDLE h = FindFirstFileW(to_wide_cstr(concat(s, make_string_c("/*"))), &data);
	if(h == INVALID_HANDLE_VALUE) 
		return;
	do
	{
		const wchar_t* local_name = data.cFileName;
		int len = WideCharToMultiByte(CP_UTF8, 0, local_name, -1, NULL, 0, NULL, NULL);
		char* dest = (char*)ALLOC(len);
		WideCharToMultiByte(CP_UTF8, 0, local_name, -1, dest, len, NULL, NULL);
		if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
		{
			add_string(&info.directory_names, 
				&info.num_directory_names, &dcap, make_string(dest, len-1));
		}
		else
		{
			add_string(&info.file_names, 
				&info.num_file_names, &fcap, make_string(dest, len-1));
		}
	}
	while(FindNextFileW(h, &data) != 0);
	memcpy(r, &info, sizeof(Directory_Info));
}
#endif
