typedef struct Image
{
	byte* pixels;
	longo width, height;
} Image;

#define STBI_MALLOC image_memory_allocate
#define STBI_REALLOC_SIZED image_memory_reallocate
#define STBI_FREE image_memory_free
static byte* image_memory;
static inta image_memory_pos;
static inta image_memory_size;
void image_set_buffer(inta size, octa buffer)
{
	image_memory = (byte*)buffer;
	image_memory_size = size;
	image_memory_pos = 0;
}
void image_set_buffer_position(inta pos)
{
	image_memory_pos = pos;
}
inta image_get_buffer_position()
{
	return image_memory_pos;
}
static void image_memory_free(void* data) {}
static void* image_memory_allocate(inta size)
{
	inta m = image_memory_pos % 8;
	if(m != 0)
		image_memory_pos += (8 - m);
	if(image_memory_pos+size > image_memory_size)
	{
		print_error("image memory allocation failed");
		return NULL;
	}
	void* r = &image_memory[image_memory_pos];
	image_memory_pos += size;
	return r;
}
static void* image_memory_reallocate(void* data, inta old_size, inta size)
{
	if(((byte*)data + old_size) == ((byte*)image_memory + image_memory_pos))
	{
		image_memory_pos += (size - old_size);
		if(image_memory_pos >= image_memory_size)
		{
			print_error("image memory reallocation failed");
			return NULL;
		}
		return data;
	}
	void* new_data = image_memory_allocate(size);
	memcpy(new_data, data, old_size);
	return new_data;
}
#define STBI_NO_FAILURE_STRINGS
#define STBI_NO_LINEAR
#define STBI_NO_HDR
#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION
#include "stb_image.h"
#include "stb_image_write.h"
void image_set_flip(byte enable)
{
	stbi_set_flip_vertically_on_load(enable);
	stbi_flip_vertically_on_write(enable);
}
void image_load(octa* r, octa* file_name_buf, tiny num_components)
{
	String file_name;
	memcpy(&file_name, file_name_buf, 16);
	Image img;
	int x, y, n;
	img.pixels = stbi_load(to_cstr(file_name), &x, &y, &n, num_components);
	img.width = x;
	img.height = y;
	memcpy(r, &img, sizeof(Image));
}
void image_load_from_memory(octa* r, inta size, octa data, tiny num_components)
{
	Image img;
	int x, y, n;
	img.pixels = stbi_load_from_memory((byte*)data, size, &x, &y, &n, num_components); 
	img.width = x;
	img.height = y;
	memcpy(r, &img, sizeof(Image));
}
byte image_save(octa* img_buf, octa* file_name_buf, tiny num_components)
{
	String file_name;
	memcpy(&file_name, file_name_buf, 16);
	Image* img = (Image*)img_buf;
	if(string_equals(string_find_extension(file_name), make_string_c("bmp")))
		return stbi_write_bmp(to_cstr(file_name), img->width, img->height,
			num_components, img->pixels) != 0;
	else if(string_equals(string_find_extension(file_name), make_string_c("png")))
		return stbi_write_png(to_cstr(file_name), img->width, img->height,
			num_components, img->pixels, img->width * num_components) != 0;
	else if(string_equals(string_find_extension(file_name), make_string_c("tga")))
		return stbi_write_tga(to_cstr(file_name), img->width, img->height,
			num_components, img->pixels) != 0;
	else if(string_equals(string_find_extension(file_name), make_string_c("jpg")))
		return stbi_write_jpg(to_cstr(file_name), img->width, img->height,
			num_components, img->pixels, 100) != 0;
	return false;
}
