#define MAX_DYNAMIC_SIZE 0x80000
static char dynamic_buffer[MAX_DYNAMIC_SIZE];
static int dynamic_pos;
static void* system_allocate(int size)
{
	int m = dynamic_pos % 8;
	if(m != 0)
		dynamic_pos += (8 - m);
	if(dynamic_pos+size > MAX_DYNAMIC_SIZE)
		dynamic_pos = 0;
	void* r = &dynamic_buffer[dynamic_pos];
	dynamic_pos += size;
	return r;
}
static void* system_reallocate(void* data, inta size, inta old_size)
{
	if(((byte*)data + old_size) == ((byte*)dynamic_buffer + dynamic_pos))
	{
		dynamic_pos += (size - old_size);
		if(dynamic_pos+size > MAX_DYNAMIC_SIZE)
		{
			memmove(dynamic_buffer, data, old_size);
			dynamic_pos = size;
			return dynamic_buffer;
		}
		return data;
	}
	void* new_data = system_allocate(size);
	memmove(new_data, data, old_size);
	return new_data;
}
SDL_Window* sdl_window;
Uint32 sdl_window_id;
SDL_Renderer* sdl_renderer;
SDL_Texture* sdl_texture;
int sdl_texture_width, sdl_texture_height;
byte sdl_vsync;
byte fullscreen;
byte opengl_initialized;
byte is_software_rendering;
SDL_GLContext opengl_context;
static void init_input();
static void init_graphics_resources();
static void init_cursors();
static void release_cursors();
void system_init(byte flags)
{
#ifdef WINDOWS_OS
	WSADATA wsaData;
	WSAStartup(MAKEWORD(1,1), &wsaData);
#endif
	quadro sdl_flags = 0;
	if(flags & INIT_TIME)
		sdl_flags |= SDL_INIT_TIMER;
	if(flags & INIT_AUDIO)
		sdl_flags |= SDL_INIT_AUDIO;
	if(flags & INIT_DISPLAY)
		sdl_flags |= SDL_INIT_VIDEO;
	if(flags & INIT_INPUT)
	{
		sdl_flags |= SDL_INIT_EVENTS;
		init_input();
	}
	if(flags & INIT_CONTROLLER)
		sdl_flags |= SDL_INIT_JOYSTICK;
	if(flags & INIT_HAPTIC)
		sdl_flags |= SDL_INIT_HAPTIC;
	SDL_Init(sdl_flags);
	init_cursors();
}
void system_release()
{
	if(sdl_texture != NULL)
	{
		SDL_DestroyTexture(sdl_texture);
		SDL_DestroyRenderer(sdl_renderer);
	}
	if(opengl_context != NULL)
		SDL_GL_DeleteContext(opengl_context);
	SDL_DestroyWindow(sdl_window);
	release_cursors();
	SDL_Quit();
#ifdef WINDOWS_OS
	WSACleanup();
#endif
}
void system_execute(octa* cmd)
{
	int result = system(to_cstr(make_string((char*)cmd[0], (inta)cmd[1])));
}
inta system_time()
{
	return time(NULL);
}
inta system_performance_counter()
{
	return SDL_GetPerformanceCounter();
}
inta system_performance_frequency()
{
	return SDL_GetPerformanceFrequency();
}
void system_sleep(inta time)
{
	SDL_Delay(time);
}
inta system_ticks()
{
	return SDL_GetTicks();
}
octa system_load_library(octa* file_name_buf)
{
	String file_name;
	memcpy(&file_name, file_name_buf, 16);
	return (octa)SDL_LoadObject(to_cstr(file_name));
}
octa system_load_function(octa h, octa* name_buf)
{
	String name;
	memcpy(&name, name_buf, 16);
	return (octa)SDL_LoadFunction((void*)h, to_cstr(name));
}
void system_unload_library(octa h)
{
	SDL_UnloadObject((void*)h);
}
static void init_sdl_renderer(longo w, longo h, byte vsync)
{
	if(sdl_renderer != NULL)
	{
		SDL_DestroyTexture(sdl_texture);
		SDL_DestroyRenderer(sdl_renderer);
	}
	sdl_texture_width = w;
	sdl_texture_height = h;
	sdl_renderer = SDL_CreateRenderer(sdl_window, -1, 
		(vsync ? SDL_RENDERER_PRESENTVSYNC : 0));
	sdl_texture = SDL_CreateTexture(sdl_renderer,
		SDL_PIXELFORMAT_ABGR8888, SDL_TEXTUREACCESS_STREAMING, w, h);
}
static void show_opengl_error()
{
	SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR,
		"Graphics not supported",
		"OpenGL 3.3 is not supported on this computer.",
		NULL);
}
static void init_opengl_renderer()
{
	if(!opengl_initialized)
	{
		if(sdl_window == NULL)
		{
			SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR,
				"Graphics error",
				"Failed to initialize hardware graphics.",
				NULL);
			exit(1);
		}
		opengl_context = SDL_GL_CreateContext(sdl_window);
		if(opengl_context == NULL)
		{
			show_opengl_error();
			exit(1);
		}
		if(!gladLoadGLLoader((GLADloadproc)SDL_GL_GetProcAddress))
		{
			show_opengl_error();
			exit(1);
		}
		opengl_initialized = true;
		init_graphics_resources();
	}
}
typedef struct Display_Info
{
	longo width, height;
	longo refresh_rate;
} Display_Info;
byte display_open(longo w, longo h, octa* title_buf, doubli flags, longo num_samples)
{
	String title;
	memcpy(&title, title_buf, 16);
	if(sdl_window != NULL)
		return false;
	Uint32 sdl_flags = 0;
	if(flags & DISPLAY_HIDDEN)
		sdl_flags |= SDL_WINDOW_HIDDEN;
	if(flags & DISPLAY_BORDERLESS)
		sdl_flags |= SDL_WINDOW_BORDERLESS;
	if(flags & DISPLAY_RESIZABLE)
		sdl_flags |= SDL_WINDOW_RESIZABLE;
	if(flags & DISPLAY_MINIMIZED)
		sdl_flags |= SDL_WINDOW_MINIMIZED;
	if(flags & DISPLAY_MAXIMIZED)
		sdl_flags |= SDL_WINDOW_MAXIMIZED;
	if(flags & DISPLAY_FULLSCREEN)
	{
		sdl_flags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
		fullscreen = true;
	}
	if((flags & DISPLAY_SOFTWARE_GRAPHICS) == 0)
		sdl_flags |= SDL_WINDOW_OPENGL;
	if(num_samples > 1)
	{
		SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
		SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, num_samples);
	}
	SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3);
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE);
	sdl_window = SDL_CreateWindow(to_cstr(title), 
		SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED,
		w, h, sdl_flags);
	sdl_window_id = SDL_GetWindowID(sdl_window);
	if(flags & DISPLAY_SOFTWARE_GRAPHICS)
	{
		is_software_rendering = true;
		init_sdl_renderer(w, h, (flags & DISPLAY_VSYNC) != 0);
	}
	else
	{
		is_software_rendering = false;
		init_opengl_renderer();
		if(flags & DISPLAY_VSYNC)
		{
			sdl_vsync = true;
			if(SDL_GL_SetSwapInterval(-1))
				SDL_GL_SetSwapInterval(1);
		}
		else
		{
			sdl_vsync = false;
			SDL_GL_SetSwapInterval(0);
		}
	}
	return sdl_window != NULL;
}
void display_close()
{
	if(sdl_window != NULL)
	{
		SDL_DestroyWindow(sdl_window);
		sdl_window = NULL;
	}
}
void display_set_size(longo w, longo h) { SDL_SetWindowSize(sdl_window, w, h); }
void display_set_title(octa* title_buf)
{
	String title;
	memcpy(&title, title_buf, 16);
	SDL_SetWindowTitle(sdl_window, to_cstr(title));
}
void display_set_icon(longo w, longo h, octa pixels)
{
	SDL_Surface* surface = SDL_CreateRGBSurfaceFrom(
		(byte*)pixels, w, h,
		32, w * 4, 0xff, 0xff00, 0xff0000, 0xff000000);
	SDL_SetWindowIcon(sdl_window, surface);
	SDL_FreeSurface(surface);
}
void display_set_content(longo w, longo h, octa pixels)
{
	if(sdl_texture == NULL || sdl_renderer == NULL)
		return;
	SDL_Rect rect;
	rect.x = 0;
	rect.y = 0;
	rect.w = w;
	rect.h = h;
	SDL_UpdateTexture(sdl_texture, &rect, (byte*)pixels, w * 4);
	SDL_RenderClear(sdl_renderer);
	SDL_RenderCopy(sdl_renderer, sdl_texture, NULL, NULL);
	SDL_RenderPresent(sdl_renderer);
}
void display_set_visible(byte visible)
{
	if(visible)
		SDL_ShowWindow(sdl_window);
	else
		SDL_HideWindow(sdl_window);
}
void display_set_border(byte border) { SDL_SetWindowBordered(sdl_window, border); }
void display_set_fullscreen(byte fullscreen)
{
	SDL_SetWindowFullscreen(sdl_window,
		fullscreen ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);
}
void display_set_vsync(byte vsync)
{
	if(is_software_rendering)
		init_sdl_renderer(sdl_texture_width, sdl_texture_height, vsync);
	else
	{
		init_opengl_renderer();
		if(vsync)
		{
			if(SDL_GL_SetSwapInterval(-1))
				SDL_GL_SetSwapInterval(1);
		}
		else
			SDL_GL_SetSwapInterval(0);
	}
}
void display_set_grab(byte grab)
{
	SDL_SetWindowGrab(sdl_window, grab);
}
void display_set_software_graphics(byte software_graphics)
{
	if(software_graphics)
		init_sdl_renderer(sdl_texture_width, sdl_texture_height, sdl_vsync);
	else
		init_opengl_renderer();
	is_software_rendering = software_graphics;
}
void display_maximize() { SDL_MaximizeWindow(sdl_window); }
void display_minimize() { SDL_MinimizeWindow(sdl_window); }
void display_restore() { SDL_RestoreWindow(sdl_window); }
void display_get_info(quadro* r)
{
	if(sdl_window == NULL)
	{
		memset(r, 0, sizeof(Display_Info));
		return;
	}
	Display_Info info;
	SDL_DisplayMode m;
	SDL_GetWindowDisplayMode(sdl_window, &m);
	info.refresh_rate = m.refresh_rate;
	if(fullscreen)
	{
		info.width = m.w;
		info.height = m.h;
	}
	else
	{
		int width, height;
		SDL_GetWindowSize(sdl_window, &width, &height);
		info.width = width;
		info.height = height;
	}
	memcpy(r, &info, sizeof(Display_Info));
}
void screensaver_enable(byte screensaver)
{
	if(screensaver)
		SDL_EnableScreenSaver();
	else
		SDL_DisableScreenSaver();
}
void show_simple_message_box(byte icon, octa* title_buf, octa* message_buf)
{
	String title, message;
	memcpy(&title, title_buf, 16);
	memcpy(&message, message_buf, 16);
	Uint32 sdl_flag;
	if(icon == MESSAGE_BOX_ICON_ERROR)
		sdl_flag = SDL_MESSAGEBOX_ERROR;
	else if(icon == MESSAGE_BOX_ICON_WARNING)
		sdl_flag = SDL_MESSAGEBOX_WARNING;
	else
		sdl_flag = SDL_MESSAGEBOX_INFORMATION;
	SDL_ShowSimpleMessageBox(sdl_flag, to_cstr(title), to_cstr(message), sdl_window);
}
static byte to_modifier_flags(Uint32 sdl_flags)
{
	byte flags = 0;
	if(sdl_flags & KMOD_CTRL)
		flags |= KEY_FLAG_CONTROL;
	if(sdl_flags & KMOD_SHIFT)
		flags |= KEY_FLAG_SHIFT;
	if(sdl_flags & KMOD_ALT)
		flags |= KEY_FLAG_ALT;
	if(sdl_flags & KMOD_GUI)
		flags |= KEY_FLAG_SUPER;
	if(sdl_flags & KMOD_CAPS)
		flags |= KEY_FLAG_CAPSLOCK;
	if(sdl_flags & KMOD_NUM)
		flags |= KEY_FLAG_NUMLOCK;
	return flags;
}
byte to_key_code[285];
static void init_key_codes()
{
	for(int i = 0; i < 285; i++)
		to_key_code[i] = 0xff;
	for(int l = 0; l < 26; l++)
		to_key_code[SDL_SCANCODE_A + l] = KEY_A + l;
	for(int d = 0; d < 10; d++)
		to_key_code[SDL_SCANCODE_1 + ((d+9) % 10)] = KEY_NUM0 + d;
	to_key_code[SDL_SCANCODE_LEFT] = KEY_LEFT;
	to_key_code[SDL_SCANCODE_RIGHT] = KEY_RIGHT;
	to_key_code[SDL_SCANCODE_UP] = KEY_UP;
	to_key_code[SDL_SCANCODE_DOWN] = KEY_DOWN;
	to_key_code[SDL_SCANCODE_SPACE] = KEY_SPACE;
	to_key_code[SDL_SCANCODE_RETURN] = KEY_ENTER;
	to_key_code[SDL_SCANCODE_BACKSPACE] = KEY_BACKSPACE;
	to_key_code[SDL_SCANCODE_TAB] = KEY_TAB;
	to_key_code[SDL_SCANCODE_ESCAPE] = KEY_ESCAPE;
	to_key_code[SDL_SCANCODE_LCTRL] = KEY_LCONTROL;
	to_key_code[SDL_SCANCODE_LSHIFT] = KEY_LSHIFT;
	to_key_code[SDL_SCANCODE_LALT] = KEY_LALT;
	to_key_code[SDL_SCANCODE_LGUI] = KEY_LSUPER;
	to_key_code[SDL_SCANCODE_RCTRL] = KEY_RCONTROL;
	to_key_code[SDL_SCANCODE_RSHIFT] = KEY_RSHIFT;
	to_key_code[SDL_SCANCODE_RALT] = KEY_RALT;
	to_key_code[SDL_SCANCODE_RGUI] = KEY_RSUPER;
	to_key_code[SDL_SCANCODE_APPLICATION] = KEY_MENU;
	to_key_code[SDL_SCANCODE_LEFTBRACKET] = KEY_LBRACKET;
	to_key_code[SDL_SCANCODE_RIGHTBRACKET] = KEY_RBRACKET;
	to_key_code[SDL_SCANCODE_SEMICOLON] = KEY_SEMICOLON;
	to_key_code[SDL_SCANCODE_COMMA] = KEY_COMMA;
	to_key_code[SDL_SCANCODE_PERIOD] = KEY_PERIOD;
	to_key_code[SDL_SCANCODE_APOSTROPHE] = KEY_QUOTE;
	to_key_code[SDL_SCANCODE_SLASH] = KEY_SLASH;
	to_key_code[SDL_SCANCODE_BACKSLASH] = KEY_BACKSLASH;
	to_key_code[SDL_SCANCODE_GRAVE] = KEY_TILDE;
	to_key_code[SDL_SCANCODE_EQUALS] = KEY_EQUAL;
	to_key_code[SDL_SCANCODE_MINUS] = KEY_HYPHEN;
	to_key_code[SDL_SCANCODE_CAPSLOCK] = KEY_CAPSLOCK;
	to_key_code[SDL_SCANCODE_NUMLOCKCLEAR] = KEY_NUMLOCK;
	to_key_code[SDL_SCANCODE_PRINTSCREEN] = KEY_PRINTSCREEN;
	to_key_code[SDL_SCANCODE_PAGEUP] = KEY_PAGE_UP;
	to_key_code[SDL_SCANCODE_PAGEDOWN] = KEY_PAGE_DOWN;
	to_key_code[SDL_SCANCODE_HOME] = KEY_HOME;
	to_key_code[SDL_SCANCODE_END] = KEY_END;
	to_key_code[SDL_SCANCODE_INSERT] = KEY_INSERT;
	to_key_code[SDL_SCANCODE_DELETE] = KEY_DELETE;
	to_key_code[SDL_SCANCODE_KP_PLUS] = KEY_ADD;
	to_key_code[SDL_SCANCODE_KP_MINUS] = KEY_SUBTRACT;
	to_key_code[SDL_SCANCODE_KP_MULTIPLY] = KEY_MULTIPLY;
	to_key_code[SDL_SCANCODE_KP_DIVIDE] = KEY_DIVIDE;
	to_key_code[SDL_SCANCODE_KP_PERIOD] = KEY_NUMPAD_PERIOD;
	to_key_code[SDL_SCANCODE_KP_ENTER] = KEY_NUMPAD_ENTER;
	to_key_code[SDL_SCANCODE_KP_0] = KEY_NUMPAD0;
	for(int d = 0; d < 9; d++)
		to_key_code[SDL_SCANCODE_KP_1 + d] = KEY_NUMPAD1 + d;
	for(int f = 0; f < 12; f++)
		to_key_code[SDL_SCANCODE_F1 + f] = KEY_F1 + f;
}
static octa find_controller_handle(SDL_JoystickID sdl_id)
{
	return (octa)SDL_JoystickFromInstanceID(sdl_id);
}
static byte keys_pressed[KEY_NUM];
static byte keys_released[KEY_NUM];
static byte keys_held[KEY_NUM];
static byte keys_typed[KEY_NUM];
static byte key_last_pressed, key_last_released;
static char text_typed[32];
static char text_edited[32];
static longo text_edited_start, text_edited_length;
static byte mouse_buttons_pressed[MOUSE_BUTTON_NUM];
static byte mouse_buttons_released[MOUSE_BUTTON_NUM];
static byte mouse_buttons_held[MOUSE_BUTTON_NUM];
static byte mouse_button_last_pressed, mouse_button_last_released;
static longo mouse_x, mouse_y;
static longo mouse_moved_x, mouse_moved_y, mouse_wheel;
static byte window_resized, window_minimized, window_maximized, window_restored;
static byte window_entered, window_left, window_focus_gained, window_focus_lost;
static byte window_closed;
static longo window_new_width, window_new_height;
static byte controller_device_added, controller_device_removed;
static longo controller_device_added_index;
static octa controller_device_removed_handle;
static char dropped_file_data[0x100];
static int dropped_file_length;
static void init_input()
{
	init_key_codes();
	memset(keys_pressed, 0, sizeof(keys_pressed));
	memset(keys_released, 0, sizeof(keys_released));
	memset(keys_held, 0, sizeof(keys_held));
	memset(keys_typed, 0, sizeof(keys_typed));
	memset(text_typed, 0, sizeof(text_typed));
	memset(text_edited, 0, sizeof(text_edited));
	text_edited_start = 0;
	text_edited_length = 0;
	memset(mouse_buttons_pressed, 0, sizeof(mouse_buttons_pressed));
	memset(mouse_buttons_released, 0, sizeof(mouse_buttons_released));
	memset(mouse_buttons_held, 0, sizeof(mouse_buttons_held));
	key_last_pressed = KEY_NONE;
	key_last_released = KEY_NONE;
	mouse_button_last_pressed = MOUSE_BUTTON_NONE;
	mouse_button_last_released = MOUSE_BUTTON_NONE;
	mouse_moved_x = 0;
	mouse_moved_y = 0;
	mouse_wheel = 0;
	window_resized = 0;
	window_minimized = 0;
	window_maximized = 0;
	window_restored = 0;
	window_entered = 0;
	window_left = 0;
	window_focus_gained = 0;
	window_focus_lost = 0;
	window_closed = 0;
	controller_device_added = 0;
	controller_device_removed = 0;
	dropped_file_length = 0;
}
static void begin_input()
{
	memset(keys_pressed, 0, sizeof(keys_pressed));
	memset(keys_released, 0, sizeof(keys_released));
	memset(keys_typed, 0, sizeof(keys_typed));
	memset(text_typed, 0, sizeof(text_typed));
	memset(text_edited, 0, sizeof(text_edited));
	text_edited_start = 0;
	text_edited_length = 0;
	memset(mouse_buttons_pressed, 0, sizeof(mouse_buttons_pressed));
	memset(mouse_buttons_released, 0, sizeof(mouse_buttons_released));
	mouse_moved_x = 0;
	mouse_moved_y = 0;
	mouse_wheel = 0;
	window_resized = 0;
	window_minimized = 0;
	window_maximized = 0;
	window_restored = 0;
	window_entered = 0;
	window_left = 0;
	window_focus_gained = 0;
	window_focus_lost = 0;
	window_closed = 0;
	controller_device_added = 0;
	controller_device_removed = 0;
	dropped_file_length = 0;
}
static byte handle_event(SDL_Event* sdl_event)
{
	switch(sdl_event->type)
	{
	case SDL_WINDOWEVENT:
		if(sdl_event->window.windowID == sdl_window_id)
		{
			switch(sdl_event->window.event)
			{
			case SDL_WINDOWEVENT_RESIZED: 
				window_new_width = sdl_event->window.data1;
				window_new_height = sdl_event->window.data2;
				window_resized = 1;
				break;
			case SDL_WINDOWEVENT_MINIMIZED: window_minimized = 1; break;
			case SDL_WINDOWEVENT_MAXIMIZED: window_maximized = 1; break;
			case SDL_WINDOWEVENT_RESTORED: window_restored = 1; break;
			case SDL_WINDOWEVENT_ENTER: window_entered = 1; break;
			case SDL_WINDOWEVENT_LEAVE: window_left = 1; break;
			case SDL_WINDOWEVENT_FOCUS_GAINED: window_focus_gained = 1; break;
			case SDL_WINDOWEVENT_FOCUS_LOST: window_focus_lost = 1; break;
			case SDL_WINDOWEVENT_CLOSE: window_closed = 1; break;
			default:
				return false;
			}
		}
		break;
	case SDL_KEYDOWN:
	case SDL_KEYUP:
	{
		byte k = to_key_code[sdl_event->key.keysym.scancode];
		if(k == 0xff)
			return false;
		if(sdl_event->type == SDL_KEYDOWN)
		{
			if(sdl_event->key.repeat)
			{
				keys_typed[KEY_ANY] = 1;
				keys_typed[k] = 1;
				if(k == KEY_LCONTROL || k == KEY_RCONTROL)
					keys_typed[KEY_CONTROL] = 1;
				else if(k == KEY_LSHIFT || k == KEY_RSHIFT)
					keys_typed[KEY_SHIFT] = 1;
				else if(k == KEY_LALT || k == KEY_RALT)
					keys_typed[KEY_ALT] = 1;
				else if(k == KEY_LSUPER || k == KEY_RSUPER)
					keys_typed[KEY_SUPER] = 1;
			}
			else
			{
				key_last_pressed = k;
				keys_typed[KEY_ANY] = 1;
				keys_typed[k] = 1;
				keys_pressed[KEY_ANY] = 1;
				keys_pressed[k] = 1;
				keys_held[KEY_ANY]++;
				keys_held[k] = 1;
				if(k == KEY_LCONTROL || k == KEY_RCONTROL)
				{
					keys_pressed[KEY_CONTROL] = 1;
					keys_typed[KEY_CONTROL] = 1;
					keys_held[KEY_CONTROL]++;
				}
				else if(k == KEY_LSHIFT || k == KEY_RSHIFT)
				{
					keys_pressed[KEY_SHIFT] = 1;
					keys_typed[KEY_SHIFT] = 1;
					keys_held[KEY_SHIFT]++;
				}
				else if(k == KEY_LALT || k == KEY_RALT)
				{
					keys_pressed[KEY_ALT] = 1;
					keys_typed[KEY_ALT] = 1;
					keys_held[KEY_ALT]++;
				}
				else if(k == KEY_LSUPER || k == KEY_RSUPER)
				{
					keys_pressed[KEY_SUPER] = 1;
					keys_typed[KEY_SUPER] = 1;
					keys_held[KEY_SUPER]++;
				}
			}
		}
		else
		{
			key_last_released = k;
			keys_released[KEY_ANY] = 1;
			keys_released[k] = 1;
			keys_held[KEY_ANY]--;
			keys_held[k] = 0;
			if(k == KEY_LCONTROL || k == KEY_RCONTROL)
			{
				keys_released[KEY_CONTROL] = 1;
				keys_held[KEY_CONTROL]--;
			}
			else if(k == KEY_LSHIFT || k == KEY_RSHIFT)
			{
				keys_released[KEY_SHIFT] = 1;
				keys_held[KEY_SHIFT]--;
			}
			else if(k == KEY_LALT || k == KEY_RALT)
			{
				keys_released[KEY_ALT] = 1;
				keys_held[KEY_ALT]--;
			}
			else if(k == KEY_LSUPER || k == KEY_RSUPER)
			{
				keys_released[KEY_SUPER] = 1;
				keys_held[KEY_SUPER]--;
			}
		}
		break;
	}
	case SDL_TEXTINPUT:
	{
		inta old_len = strlen(text_typed);
		char* text = sdl_event->text.text;
		inta new_len = strlen(text)+1;
		if(old_len + new_len > sizeof(text_typed))
			old_len = 0; //clear old text if it doesn't fit...
		memcpy(&text_typed[old_len], text, new_len);
		break;
	}
	case SDL_TEXTEDITING:
	{
		text_edited_start = sdl_event->edit.start;
		text_edited_length = sdl_event->edit.length;
		char* text = sdl_event->edit.text;
		memcpy(text_edited, text, strlen(text)+1);
		break;
	}
	case SDL_MOUSEMOTION:
		mouse_x = sdl_event->motion.x;
		mouse_y = sdl_event->motion.y;
		mouse_moved_x += sdl_event->motion.xrel;
		mouse_moved_y += sdl_event->motion.yrel;
		break;
	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
	{
		int button = MOUSE_BUTTON_NONE;
		switch(sdl_event->button.button)
		{
		case SDL_BUTTON_LEFT:
			button = MOUSE_BUTTON_LEFT; break;
		case SDL_BUTTON_MIDDLE:
			button = MOUSE_BUTTON_MIDDLE; break;
		case SDL_BUTTON_RIGHT:
			button = MOUSE_BUTTON_RIGHT; break;
		default:
			return false;
		}
		if(sdl_event->type == SDL_MOUSEBUTTONDOWN)
		{
			mouse_button_last_pressed = button;
			mouse_buttons_pressed[MOUSE_BUTTON_ANY] = 1;
			mouse_buttons_pressed[button] = 1;
			mouse_buttons_held[MOUSE_BUTTON_ANY]++;
			mouse_buttons_held[button] = 1;
		}
		else
		{
			mouse_button_last_released = button;
			mouse_buttons_released[MOUSE_BUTTON_ANY] = 1;
			mouse_buttons_released[button] = 1;
			mouse_buttons_held[MOUSE_BUTTON_ANY]--;
			mouse_buttons_held[button] = 0;
		}
		break;
	}
	case SDL_MOUSEWHEEL: mouse_wheel += sdl_event->wheel.y; break;
	case SDL_JOYDEVICEADDED:
		controller_device_added = 1;
		controller_device_added_index = sdl_event->jdevice.which;
		break;
	case SDL_JOYDEVICEREMOVED:
		controller_device_removed = 1;
		controller_device_removed_handle = find_controller_handle(sdl_event->jdevice.which);		
		break;
	case SDL_DROPFILE:
	{
		int len = strlen(sdl_event->drop.file);
		if(len > sizeof(dropped_file_data))
			len = sizeof(dropped_file_data);
		memcpy(dropped_file_data, sdl_event->drop.file, len);
		dropped_file_length = len;
		SDL_free(sdl_event->drop.file);
		break;
	}
	case SDL_QUIT:
		window_closed = 1;
		return true;
	default:
		return false;
	}
	return true;
}
void input_poll()
{
	begin_input();
	SDL_Event event;
	while(SDL_PollEvent(&event))
		handle_event(&event);
}
byte display_is_minimized() { return window_minimized; }
byte display_is_maximized() { return window_maximized; }
byte display_is_restored() { return window_restored; }
byte display_is_entered() { return window_entered; }
byte display_is_left() { return window_left; }
byte display_is_focus_gained() { return window_focus_gained; }
byte display_is_focus_lost() { return window_focus_lost; }
byte display_is_closed() { return window_closed; }
byte display_is_resized(longo* new_width, longo* new_height)
{
	if(window_resized)
	{
		*new_width = window_new_width;
		*new_height = window_new_height;
		return true;
	}
	return false;
}
void display_file_dropped(octa* r)
{
	r[0] = (octa)dropped_file_data;
	r[1] = dropped_file_length;
}
byte input_last_pressed_key() { return key_last_pressed; }
byte input_last_released_key() { return key_last_released; }
byte input_last_pressed_mouse_button() { return mouse_button_last_pressed; }
byte input_last_released_mouse_button() { return mouse_button_last_released; }
byte input_key_is_pressed(byte key) { return keys_pressed[key]; }
byte input_key_is_released(byte key) { return keys_released[key]; }
byte input_key_is_held(byte key) { return keys_held[key]; }
byte input_key_is_typed(byte key) { return keys_typed[key]; }
void input_key_held_state(octa* r)
{
	r[0] = (octa)keys_held;
	r[1] = KEY_NUM;
}
void input_key_pressed_state(octa* r)
{
	r[0] = (octa)keys_pressed;
	r[1] = KEY_NUM;
}
void input_key_released_state(octa* r)
{
	r[0] = (octa)keys_released;
	r[1] = KEY_NUM;
}
byte input_key_modifiers() { return to_modifier_flags(SDL_GetModState()); }
void input_text_enable(byte enable)
{
	if(enable)
		SDL_StartTextInput();
	else
		SDL_StopTextInput();
}
void input_text_rect(longo x, longo y, longo w, longo h)
{
	SDL_Rect rect;
	rect.x = x;
	rect.y = y;
	rect.w = w;
	rect.h = h;
	SDL_SetTextInputRect(&rect);
}
void input_text_typed(octa* r)
{
	octa len = strlen(text_typed);
	r[0] = (octa)text_typed;
	r[1] = len;
}
void input_text_edited(octa* r, longo* start, longo* length)
{
	if(start != NULL) *start = text_edited_start;
	if(length != NULL) *length = text_edited_length;
	octa len = strlen(text_edited);
	r[0] = (octa)text_edited;
	r[1] = len;
}
byte input_mouse_is_pressed(byte button)
{
	return mouse_buttons_pressed[button];
}
byte input_mouse_is_released(byte button)
{
	return mouse_buttons_released[button];
}
byte input_mouse_is_held(byte button)
{
	return mouse_buttons_held[button];
}
longo input_mouse_wheel()
{
	return mouse_wheel;
}
void input_mouse_moved(longo* x, longo* y)
{
	*x = mouse_moved_x;
	*y = mouse_moved_y;
}
void input_mouse_held_state(octa* r)
{
	r[0] = (octa)mouse_buttons_held;
	r[1] = MOUSE_BUTTON_NUM;
}
void input_mouse_pressed_state(octa* r)
{
	r[0] = (octa)mouse_buttons_pressed;
	r[1] = MOUSE_BUTTON_NUM;
}
void input_mouse_released_state(octa* r)
{
	r[0] = (octa)mouse_buttons_released;
	r[1] = MOUSE_BUTTON_NUM;
}
void input_mouse_position(longo* x, longo* y)
{
	*x = mouse_x;
	*y = mouse_y;
}
void input_mouse_global_position(longo* state_x, longo* state_y)
{
	int x, y;
	SDL_GetGlobalMouseState(&x, &y);
	*state_x = x;
	*state_y = y;
}
void input_mouse_set_visible(byte visible)
{
	SDL_ShowCursor(visible ? SDL_ENABLE : SDL_DISABLE);
}
void input_mouse_set_relative(byte relative)
{
	SDL_SetRelativeMouseMode(relative ? SDL_TRUE : SDL_FALSE);
}
void input_mouse_set_position(longo x, longo y)
{
	SDL_WarpMouseInWindow(sdl_window, x, y);
}
void input_mouse_set_global_position(longo x, longo y)
{
	SDL_WarpMouseGlobal(x, y);
}
SDL_Cursor* system_cursors[CURSOR_NUM];
static void init_cursors()
{
	system_cursors[CURSOR_ARROW] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_ARROW);
	system_cursors[CURSOR_IBEAM] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_IBEAM);
	system_cursors[CURSOR_WAIT] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_WAIT);
	system_cursors[CURSOR_CROSSHAIR] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_CROSSHAIR);
	system_cursors[CURSOR_WAITARROW] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_WAITARROW);
	system_cursors[CURSOR_SIZENWSE] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENWSE);
	system_cursors[CURSOR_SIZENESW] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENESW);
	system_cursors[CURSOR_SIZEWE] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZEWE);
	system_cursors[CURSOR_SIZENS] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZENS);
	system_cursors[CURSOR_SIZEALL] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_SIZEALL);
	system_cursors[CURSOR_NO] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_NO);
	system_cursors[CURSOR_HAND] = SDL_CreateSystemCursor(SDL_SYSTEM_CURSOR_HAND);
}
static void release_cursors()
{
	for(int i = 0; i < CURSOR_NUM; i++)
		if(system_cursors[i])
			SDL_FreeCursor(system_cursors[i]);
}
void input_mouse_set_cursor(byte cursor) { SDL_SetCursor(system_cursors[cursor]); }
byte input_mouse_is_haptic() { return SDL_MouseIsHaptic(); }
byte controller_is_added(longo* index)
{
	if(index != NULL) *index = controller_device_added_index;
	return controller_device_added;
}
byte controller_is_removed(octa* controller)
{
	if(controller != NULL) *controller = controller_device_removed_handle;
	return controller_device_removed;
}
inta controller_num()
{
	return SDL_NumJoysticks();
}
void controller_name_for_index(octa* r, longo index)
{
	String name = make_string_c(SDL_JoystickNameForIndex(index));
	memcpy(r, &name, 16);
}
octa controller_open(longo index) { return (octa)SDL_JoystickOpen(index); }
void controller_close(octa c) { SDL_JoystickClose((SDL_Joystick*)c); }
void controller_name(octa* r, octa c)
{
	String name = make_string_c(SDL_JoystickName((SDL_Joystick*)c));
	memcpy(r, &name, 16);
}
longo controller_num_buttons(octa c) { return SDL_JoystickNumButtons((SDL_Joystick*)c); }
longo controller_num_axes(octa c) { return SDL_JoystickNumAxes((SDL_Joystick*)c); }
longo controller_num_balls(octa c) { return SDL_JoystickNumBalls((SDL_Joystick*)c); }
longo controller_num_hats(octa c) { return SDL_JoystickNumHats((SDL_Joystick*)c); }
byte controller_is_haptic(octa c) { return SDL_JoystickIsHaptic((SDL_Joystick*)c); }
byte controller_is_connected(octa c) { return SDL_JoystickGetAttached((SDL_Joystick*)c); }
byte controller_power(octa c)
{
	byte power = 0;
	Uint32 sdl_power = SDL_JoystickCurrentPowerLevel((SDL_Joystick*)c);
	if(sdl_power == SDL_JOYSTICK_POWER_UNKNOWN)
		power = CONTROLLER_POWER_UNKNOWN;
	if(sdl_power == SDL_JOYSTICK_POWER_EMPTY)
		power = CONTROLLER_POWER_EMPTY;
	if(sdl_power == SDL_JOYSTICK_POWER_LOW)
		power = CONTROLLER_POWER_LOW;
	if(sdl_power == SDL_JOYSTICK_POWER_MEDIUM)
		power = CONTROLLER_POWER_MEDIUM;
	if(sdl_power == SDL_JOYSTICK_POWER_FULL)
		power = CONTROLLER_POWER_FULL;
	if(sdl_power == SDL_JOYSTICK_POWER_WIRED)
		power = CONTROLLER_POWER_WIRED;
	if(sdl_power == SDL_JOYSTICK_POWER_MAX)
		power = CONTROLLER_POWER_MAX;
	return power;
}
byte controller_button(octa c, longo button) { return SDL_JoystickGetButton((SDL_Joystick*)c, button); }
shorti controller_axis(octa c, longo axis) { return SDL_JoystickGetAxis((SDL_Joystick*)c, axis); }
void controller_ball(octa c, longo ball, longo* x, longo* y) { SDL_JoystickGetBall((SDL_Joystick*)c, ball, x, y); }
byte controller_hat(octa c, longo hat) { return SDL_JoystickGetHat((SDL_Joystick*)c, hat); }
octa haptic_open_controller(octa c) { return (octa)SDL_HapticOpenFromJoystick((SDL_Joystick*)c); }
octa haptic_open_mouse() { return (octa)SDL_HapticOpenFromMouse(); }
void haptic_close(octa h) { SDL_HapticClose((SDL_Haptic*)h); }
byte haptic_has_rumble(octa h) { return SDL_HapticRumbleSupported((SDL_Haptic*)h); }
byte haptic_init_rumble(octa h) { return SDL_HapticRumbleInit((SDL_Haptic*)h) == 0; }
void haptic_play_rumble(octa h, quadro strength, inta length)
{
	SDL_HapticRumblePlay((SDL_Haptic*)h, *(float*)&strength, length);
}
void haptic_stop_rumble(octa h)
{
	SDL_HapticRumbleStop((SDL_Haptic*)h);
}
static char* clipboard_buffer = NULL;
static inta clipboard_memory_pos;
static inta clipboard_memory_size;
void clipboard_set_buffer(inta size, octa data)
{
	clipboard_buffer = (char*)data;
	clipboard_memory_size = size;
	clipboard_memory_pos = 0;
}
void clipboard_set_buffer_position(inta pos)
{
	clipboard_memory_pos = pos;
}
inta clipboard_get_buffer_position()
{
	return clipboard_memory_pos;
}
static char* clipboard_memory_allocate(inta size)
{
	if(image_memory_pos+size > image_memory_size)
	{
		print_error("clipboard memory allocation failed");
		return NULL;
	}
	char* r = clipboard_buffer + clipboard_memory_pos;
	clipboard_memory_pos += size;
	return r;
}
void clipboard_get(octa* r)
{
	String text = empty_string();
	if(SDL_HasClipboardText())
	{
		char* sdl_text = SDL_GetClipboardText();
		inta len = strlen(sdl_text);
		char* memory;
		if(clipboard_buffer == NULL)
			memory = ALLOC(len);
		else
			memory = clipboard_memory_allocate(len);
		memcpy(memory, sdl_text, len);
		text = make_string(memory, len);
		SDL_free(sdl_text);
	}
	memcpy(r, &text, 16);
}
void clipboard_set(octa* sb)
{
	String s;
	memcpy(&s, sb, 16);
	SDL_SetClipboardText(to_cstr(s));
}
static longo gl_format(longo format)
{
	switch(format)
	{
	case FORMAT_R8: return GL_R8;
	case FORMAT_R16: return GL_R16;
	case FORMAT_RG8: return GL_RG8;
	case FORMAT_RG16: return GL_RG16;
	case FORMAT_RGB8: return GL_RGB8;
	case FORMAT_RGB16: return GL_RGB16;
	case FORMAT_RGBA8: return GL_RGBA8;
	case FORMAT_RGBA16: return GL_RGBA16;
	case FORMAT_R8S: return GL_R8_SNORM;
	case FORMAT_R16S: return GL_R16_SNORM;
	case FORMAT_RG8S: return GL_RG8_SNORM;
	case FORMAT_RG16S: return GL_RG16_SNORM;
	case FORMAT_RGB8S: return GL_RGB8_SNORM;
	case FORMAT_RGB16S: return GL_RGB16_SNORM;
	case FORMAT_RGBA8S: return GL_RGBA8_SNORM;
	case FORMAT_RGBA16S: return GL_RGBA16_SNORM;
	case FORMAT_R16F: return GL_R16F;
	case FORMAT_R32F: return GL_R32F;
	case FORMAT_RG16F: return GL_RG16F;
	case FORMAT_RG32F: return GL_RG32F;
	case FORMAT_RGB16F: return GL_RGB16F;
	case FORMAT_RGB32F: return GL_RGB32F;
	case FORMAT_RGBA16F: return GL_RGBA16F;
	case FORMAT_RGBA32F: return GL_RGBA32F;
	case FORMAT_R8I: return GL_R8I;
	case FORMAT_R16I: return GL_R16I;
	case FORMAT_R32I: return GL_R32I;
	case FORMAT_RG8I: return GL_RG8I;
	case FORMAT_RG16I: return GL_RG16I;
	case FORMAT_RG32I: return GL_RG32I;
	case FORMAT_RGB8I: return GL_RGB8I;
	case FORMAT_RGB16I: return GL_RGB16I;
	case FORMAT_RGB32I: return GL_RGB32I;
	case FORMAT_RGBA8I: return GL_RGBA8I;
	case FORMAT_RGBA16I: return GL_RGBA16I;
	case FORMAT_RGBA32I: return GL_RGBA32I;
	case FORMAT_R8Q: return GL_R8UI;
	case FORMAT_R16Q: return GL_R16UI;
	case FORMAT_R32Q: return GL_R32UI;
	case FORMAT_RG8Q: return GL_RG8UI;
	case FORMAT_RG16Q: return GL_RG16UI;
	case FORMAT_RG32Q: return GL_RG32UI;
	case FORMAT_RGB8Q: return GL_RGB8UI;
	case FORMAT_RGB16Q: return GL_RGB16UI;
	case FORMAT_RGB32Q: return GL_RGB32UI;
	case FORMAT_RGBA8Q: return GL_RGBA8UI;
	case FORMAT_RGBA16Q: return GL_RGBA16UI;
	case FORMAT_RGBA32Q: return GL_RGBA32UI;
	case FORMAT_DEPTH16: return GL_DEPTH_COMPONENT16;
	case FORMAT_DEPTH24: return GL_DEPTH_COMPONENT24;
	case FORMAT_DEPTH32: return GL_DEPTH_COMPONENT32;
	case FORMAT_DEPTH32F: return GL_DEPTH_COMPONENT32F;
	case FORMAT_STENCIL8: return GL_STENCIL_INDEX8;
	case FORMAT_DEPTH24_STENCIL8: return GL_DEPTH24_STENCIL8;
	case FORMAT_DEPTH32F_STENCIL8: return GL_DEPTH32F_STENCIL8;
	case FORMAT_R3_G3_B2: return GL_R3_G3_B2;
	case FORMAT_RGB4: return GL_RGB4;
	case FORMAT_RGBA4: return GL_RGBA4;
	case FORMAT_RGB5: return GL_RGB5;
	case FORMAT_RGB5_A1: return GL_RGB5_A1;
	case FORMAT_RGB10: return GL_RGB10;
	case FORMAT_RGB10_A2: return GL_RGB10_A2;
	case FORMAT_RGB10_A2Q: return GL_RGB10_A2UI;
	case FORMAT_RGB12: return GL_RGB12;
	case FORMAT_RGBA12: return GL_RGBA12;
	case FORMAT_R11F_G11F_B10F: return GL_R11F_G11F_B10F;
	case FORMAT_RGB9_E5: return GL_RGB9_E5;
	case FORMAT_SRGB8: return GL_SRGB8;
	case FORMAT_SRGBA8: return GL_SRGB8_ALPHA8;
	case FORMAT_COMPRESSED_R: return GL_COMPRESSED_RED;
	case FORMAT_COMPRESSED_RG: return GL_COMPRESSED_RG;
	case FORMAT_COMPRESSED_RGB: return GL_COMPRESSED_RGB;
	case FORMAT_COMPRESSED_RGBA: return GL_COMPRESSED_RGBA;
	case FORMAT_COMPRESSED_SRGB: return GL_COMPRESSED_SRGB;
	case FORMAT_COMPRESSED_SRGBA: return GL_COMPRESSED_SRGB_ALPHA;
	case FORMAT_COMPRESSED_R_RGTC1: return GL_COMPRESSED_RED_RGTC1;
	case FORMAT_COMPRESSED_SIGNED_R_RGTC1: return GL_COMPRESSED_SIGNED_RED_RGTC1;
	case FORMAT_COMPRESSED_RG_RGTC2: return GL_COMPRESSED_RG_RGTC2;
	case FORMAT_COMPRESSED_SIGNED_RG_RGTC2: return GL_COMPRESSED_SIGNED_RG_RGTC2;
	case FORMAT_COMPRESSED_RGB_S3TC_DXT1: return GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
	case FORMAT_COMPRESSED_RGBA_S3TC_DXT1: return GL_COMPRESSED_RGBA_S3TC_DXT1_EXT;
	case FORMAT_COMPRESSED_RGBA_S3TC_DXT3: return GL_COMPRESSED_RGBA_S3TC_DXT3_EXT;
	case FORMAT_COMPRESSED_RGBA_S3TC_DXT5: return GL_COMPRESSED_RGBA_S3TC_DXT5_EXT;
	}
	return 0;
}
static longo gl_components(longo components)
{
	switch(components)
	{
	case COMPONENTS_DEPTH: return GL_DEPTH_COMPONENT;
	case COMPONENTS_STENCIL: return GL_STENCIL_INDEX;
	case COMPONENTS_DEPTH_STENCIL: return GL_DEPTH_STENCIL;
	case COMPONENTS_RED: return GL_RED;
	case COMPONENTS_GREEN: return GL_GREEN;
	case COMPONENTS_BLUE: return GL_BLUE;
	case COMPONENTS_RG: return GL_RG;
	case COMPONENTS_RGB: return GL_RGB;
	case COMPONENTS_RGBA: return GL_RGBA;
	case COMPONENTS_RED_INTEGER: return GL_RED_INTEGER;
	case COMPONENTS_GREEN_INTEGER: return GL_GREEN_INTEGER;
	case COMPONENTS_BLUE_INTEGER: return GL_BLUE_INTEGER;
	case COMPONENTS_RG_INTEGER: return GL_RG_INTEGER;
	case COMPONENTS_RGB_INTEGER: return GL_RGB_INTEGER;
	case COMPONENTS_RGBA_INTEGER: return GL_RGBA_INTEGER;
	}
	return 0;
}
static longo gl_data_type(longo data_type)
{
	switch(data_type)
	{
	case TYPE_BYTE: return GL_UNSIGNED_BYTE;
	case TYPE_DOUBLE: return GL_UNSIGNED_SHORT;
	case TYPE_QUADRO: return GL_UNSIGNED_INT;
	case TYPE_TINY: return GL_BYTE;
	case TYPE_SHORT: return GL_SHORT;
	case TYPE_LONG: return GL_INT;
	case TYPE_FLOAT: return GL_FLOAT;
	case TYPE_BYTE_3_3_2: return GL_UNSIGNED_BYTE_3_3_2;
	case TYPE_DOUBLE_4_4_4_4: return GL_UNSIGNED_SHORT_4_4_4_4;
	case TYPE_DOUBLE_5_5_5_1: return GL_UNSIGNED_SHORT_5_5_5_1;
	case TYPE_QUADRO_8_8_8_8: return GL_UNSIGNED_INT_8_8_8_8;
	case TYPE_QUADRO_10_10_10_2: return GL_UNSIGNED_INT_10_10_10_2;
	case TYPE_QUADRO_24_8: return GL_UNSIGNED_INT_24_8;
	case TYPE_FLOAT_32_QUADRO_24_8: return GL_FLOAT_32_UNSIGNED_INT_24_8_REV;
	}
	return 0;
}
static longo gl_tex_target(longo target)
{
	switch(target)
	{
	case TEXTURE_1D: return GL_TEXTURE_1D;
	case TEXTURE_2D: return GL_TEXTURE_2D;
	case TEXTURE_3D: return GL_TEXTURE_3D;
	case TEXTURE_CUBE: return GL_TEXTURE_CUBE_MAP;
	case TEXTURE_BUFFER: return GL_TEXTURE_BUFFER;
	case TEXTURE_1D_ARRAY: return GL_TEXTURE_1D_ARRAY;
	case TEXTURE_2D_ARRAY: return GL_TEXTURE_2D_ARRAY;
	case TEXTURE_2D_MULTISAMPLE: return GL_TEXTURE_2D_MULTISAMPLE;
	case TEXTURE_2D_MULTISAMPLE_ARRAY: return GL_TEXTURE_2D_MULTISAMPLE_ARRAY;
	}
	return 0;
}
static longo gl_wrap(longo wrap)
{
	switch(wrap)
	{
	case WRAP_REPEAT: return GL_REPEAT;
	case WRAP_MIRRORED_REPEAT: return GL_MIRRORED_REPEAT;
	case WRAP_CLAMP_TO_EDGE: return GL_CLAMP_TO_EDGE;
	case WRAP_CLAMP_TO_BORDER: return GL_CLAMP_TO_BORDER;
	}
	return 0;
}
static longo gl_filter(longo filter)
{
	switch(filter)
	{
	case FILTER_NEAREST: return GL_NEAREST;
	case FILTER_LINEAR: return GL_LINEAR;
	case FILTER_NEAREST_MIPMAP_NEAREST: return GL_NEAREST_MIPMAP_NEAREST;
	case FILTER_LINEAR_MIPMAP_NEAREST: return GL_LINEAR_MIPMAP_NEAREST;
	case FILTER_NEAREST_MIPMAP_LINEAR: return GL_NEAREST_MIPMAP_LINEAR;
	case FILTER_LINEAR_MIPMAP_LINEAR: return GL_LINEAR_MIPMAP_LINEAR;
	}
	return 0;
}
static longo gl_swizzle(longo swizzle)
{
	switch(swizzle)
	{
	case SWIZZLE_ZERO: return GL_ZERO;
	case SWIZZLE_ONE: return GL_ONE;
	case SWIZZLE_RED: return GL_RED;
	case SWIZZLE_GREEN: return GL_GREEN;
	case SWIZZLE_BLUE: return GL_BLUE;
	case SWIZZLE_ALPHA: return GL_ALPHA;
	}
	return 0;
}
static longo gl_tex_param(longo param)
{
	switch(param)
	{
	case TEXTURE_PARAMETER_WRAP_X: return GL_TEXTURE_WRAP_S;
	case TEXTURE_PARAMETER_WRAP_Y: return GL_TEXTURE_WRAP_T;
	case TEXTURE_PARAMETER_WRAP_Z: return GL_TEXTURE_WRAP_R;
	case TEXTURE_PARAMETER_MIN_FILTER: return GL_TEXTURE_MIN_FILTER;
	case TEXTURE_PARAMETER_MAG_FILTER: return GL_TEXTURE_MAG_FILTER;
	case TEXTURE_PARAMETER_BORDER: return GL_TEXTURE_BORDER_COLOR;
	case TEXTURE_PARAMETER_BASE_LEVEL: return GL_TEXTURE_BASE_LEVEL;
	case TEXTURE_PARAMETER_MAX_LEVEL: return GL_TEXTURE_MAX_LEVEL;
	case TEXTURE_PARAMETER_MIN_LOD: return GL_TEXTURE_MIN_LOD;
	case TEXTURE_PARAMETER_MAX_LOD: return GL_TEXTURE_MAX_LOD;
	case TEXTURE_PARAMETER_MAX_ANISOTROPY: return GL_TEXTURE_MAX_ANISOTROPY_EXT;
	}
	return 0;
}
//textures
quadro texture_create() { quadro id; glGenTextures(1, &id); return id; }
void texture_destroy(quadro id) { glDeleteTextures(1, &id); }
void texture_bind(longo target, quadro id) { glBindTexture(gl_tex_target(target), id); }
void texture_unbind(longo target) { glBindTexture(gl_tex_target(target), 0); }
void texture_unit(longo unit) { glActiveTexture(GL_TEXTURE0 + unit); }
void texture_image_1d(longo target, longo level, longo format, longo width, longo components, longo data_type, octa data)
{
	glTexImage1D(gl_tex_target(target), level, gl_format(format), width, 0, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_image_2d(longo target, longo level, longo format, longo width, longo height, longo components, longo data_type, octa data)
{
	glTexImage2D(gl_tex_target(target), level, gl_format(format), width, height, 0, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_image_3d(longo target, longo level, longo format, longo width, longo height, longo depth, longo components, longo data_type, octa data)
{
	glTexImage3D(gl_tex_target(target), level, gl_format(format), width, height, depth, 0, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_image_buffer(longo target, longo format, quadro buffer_id)
{
	glTexBuffer(gl_tex_target(target), gl_format(format), buffer_id);
}
void texture_image_2d_multisample(longo target, longo samples, longo format, longo width, longo height, byte fixed_sample_locations)
{
	glTexImage2DMultisample(gl_tex_target(target), samples, gl_format(format), width, height, fixed_sample_locations);
}
void texture_image_2d_multisample_array(longo target, longo samples, longo format, longo width, longo height, longo depth, byte fixed_sample_locations)
{
	glTexImage3DMultisample(gl_tex_target(target), samples, gl_format(format), width, height, depth, fixed_sample_locations);
}
void texture_sub_image_1d(longo target, longo level, longo x, longo width, longo components, longo data_type, octa data)
{
	glTexSubImage1D(gl_tex_target(target), level, x, width, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_sub_image_2d(longo target, longo level, longo x, longo y, longo width, longo height, longo components, longo data_type, octa data)
{
	glTexSubImage2D(gl_tex_target(target), level, x, y, width, height, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_sub_image_3d(longo target, longo level, longo x, longo y, longo z, longo width, longo height, longo depth, longo components, longo data_type, octa data)
{
	glTexSubImage3D(gl_tex_target(target), level, x, y, z, width, height, depth, gl_components(components), gl_data_type(data_type), (void*)data);
}
void texture_compressed_image_1d(longo target, longo level, longo format, longo width, inta size, octa data)
{
	glCompressedTexImage1D(gl_tex_target(target), level, gl_format(format), width, 0, size, (void*)data);
}
void texture_compressed_image_2d(longo target, longo level, longo format, longo width, longo height, inta size, octa data)
{
	glCompressedTexImage2D(gl_tex_target(target), level, gl_format(format), width, height, 0, size, (void*)data);
}
void texture_compressed_image_3d(longo target, longo level, longo format, longo width, longo height, longo depth, inta size, octa data)
{
	glCompressedTexImage3D(gl_tex_target(target), level, gl_format(format), width, height, depth, 0, size, (void*)data);
}
void texture_compressed_sub_image_1d(longo target, longo level, longo x, longo width, longo format, inta size, octa data)
{
	glCompressedTexSubImage1D(gl_tex_target(target), level, x, width, gl_format(format), size, (void*)data);
}
void texture_compressed_sub_image_2d(longo target, longo level, longo x, longo y, longo width, longo height, longo format, inta size, octa data)
{
	glCompressedTexSubImage2D(gl_tex_target(target), level, x, y, width, height, gl_format(format), size, (void*)data);
}
void texture_compressed_sub_image_3d(longo target, longo level, longo x, longo y, longo z, longo width, longo height, longo depth, longo format, inta size, octa data)
{
	glCompressedTexSubImage3D(gl_tex_target(target), level, x, y, z, width, height, depth, gl_format(format), size, (void*)data);
}
void texture_copy_image_1d(longo target, longo level, longo format, longo x, longo y, longo width)
{
	glCopyTexImage1D(gl_tex_target(target), level, gl_format(format), x, y, width, 0);
}
void texture_copy_image_2d(longo target, longo level, longo format, longo x, longo y, longo width, longo height)
{	
	glCopyTexImage2D(gl_tex_target(target), level, gl_format(format), x, y, width, height, 0);
}
void texture_copy_sub_image_1d(longo target, longo level, longo src_x, longo src_y, longo dst_x, longo width)
{
	glCopyTexSubImage1D(gl_tex_target(target), level, dst_x, src_x, src_y, width);
}
void texture_copy_sub_image_2d(longo target, longo level, longo src_x, longo src_y, longo dst_x, longo dst_y, longo width, longo height)
{
	glCopyTexSubImage2D(gl_tex_target(target), level, dst_x, dst_y, src_x, src_y, width, height);
}
void texture_copy_sub_image_3d(longo target, longo level, longo src_x, longo src_y, longo dst_x, longo dst_y, longo dst_z, longo width, longo height)
{
	glCopyTexSubImage3D(gl_tex_target(target), level, dst_x, dst_y, dst_z, src_x, src_y, width, height);
}
void texture_parameter_wrap(longo target, longo parameter, longo wrap)
{
	glTexParameteri(gl_tex_target(target), gl_tex_param(parameter), gl_wrap(wrap));
}
void texture_parameter_filter(longo target, longo parameter, longo filter)
{
	glTexParameteri(gl_tex_target(target), gl_tex_param(parameter), gl_filter(filter));
}
void texture_parameter_swizzle(longo target, longo r, longo g, longo b, longo a)
{
	GLint rgba[] = { gl_swizzle(r), gl_swizzle(g), gl_swizzle(b), gl_swizzle(a) };
	glTexParameteriv(gl_tex_target(target), GL_TEXTURE_SWIZZLE_RGBA, rgba);
}
void texture_parameterf(longo target, longo parameter, quadro value)
{
	glTexParameterf(gl_tex_target(target), gl_tex_param(parameter), *(float*)(&value));
}
void texture_parameterfa(longo target, longo parameter, octa* values)
{
	glTexParameterfv(gl_tex_target(target), gl_tex_param(parameter), (float*)values[0]);
}
void texture_parameteri(longo target, longo parameter, longo value)
{
	glTexParameteri(gl_tex_target(target), gl_tex_param(parameter), value);
}
void texture_parameteria(longo target, longo parameter, octa* values)
{
	glTexParameteriv(gl_tex_target(target), gl_tex_param(parameter), (longo*)values[0]);
}
void texture_integer_parameteria(longo target, longo parameter, octa* values)
{
	glTexParameterIiv(gl_tex_target(target), gl_tex_param(parameter), (longo*)values[0]);
}
void texture_integer_parameterqa(longo target, longo parameter, octa* values)
{
	glTexParameterIuiv(gl_tex_target(target), gl_tex_param(parameter), (quadro*)values[0]);
}
void texture_generate_mipmap(longo target) { glGenerateMipmap(gl_tex_target(target)); }
void texture_get_image(longo target, longo level, longo components, longo data_type, octa data)
{
	glGetTexImage(gl_tex_target(target), level, gl_components(components), gl_data_type(data_type), (void*)data);
}
//frame_buffer
static longo gl_fb_target(longo target)
{
	switch(target)
	{
	case FRAME_BUFFER_DRAW: return GL_DRAW_FRAMEBUFFER;
	case FRAME_BUFFER_READ: return GL_READ_FRAMEBUFFER;
	case FRAME_BUFFER_DRAW_READ: return GL_FRAMEBUFFER;
	}
	return 0;
}
static longo gl_attachment(longo attachment, longo buffer_index)
{
	switch(attachment)
	{
	case ATTACHMENT_DEPTH: return GL_DEPTH_ATTACHMENT;
	case ATTACHMENT_STENCIL: return GL_STENCIL_ATTACHMENT;
	case ATTACHMENT_DEPTH_STENCIL: return GL_DEPTH_STENCIL_ATTACHMENT;
	case ATTACHMENT_COLOR: return GL_COLOR_ATTACHMENT0 + buffer_index;
	}
	return 0;
}
static quadro gl_buffer_flags(byte flags)
{
	quadro gl_flags = 0;
	if(flags & BUFFER_FLAG_COLOR)
		gl_flags |= GL_COLOR_BUFFER_BIT;
	if(flags & BUFFER_FLAG_DEPTH)
		gl_flags |= GL_DEPTH_BUFFER_BIT;
	if(flags & BUFFER_FLAG_STENCIL)
		gl_flags |= GL_STENCIL_BUFFER_BIT;
	return gl_flags;
}
quadro frame_buffer_create() { quadro id; glGenFramebuffers(1, &id); return id; }
void frame_buffer_destroy(quadro id) { glDeleteFramebuffers(1, &id); }
void frame_buffer_bind(longo target, quadro id) { glBindFramebuffer(gl_fb_target(target), id); }
void frame_buffer_unbind(longo target) { glBindFramebuffer(gl_fb_target(target), 0); }
void frame_buffer_texture(longo target, longo attachment, longo buffer_index, quadro texture_id, longo level)
{
	glFramebufferTexture(gl_fb_target(target), gl_attachment(attachment, buffer_index), texture_id, level);
}
void frame_buffer_texture_face(longo target, longo attachment, longo buffer_index, quadro texture_id, longo level, longo face)
{
	glFramebufferTexture2D(gl_fb_target(target), gl_attachment(attachment, buffer_index), GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, texture_id, level);
}
void frame_buffer_texture_layer(longo target, longo attachment, longo buffer_index, quadro texture_id, longo level, longo layer)
{
	glFramebufferTextureLayer(gl_fb_target(target), gl_attachment(attachment, buffer_index), texture_id, level, layer);
}
void frame_buffer_render_buffer(longo target, longo attachment, longo buffer_index, quadro render_buffer_id)
{
	glFramebufferRenderbuffer(gl_fb_target(target), gl_attachment(attachment, buffer_index), GL_RENDERBUFFER, render_buffer_id);
}
longo frame_buffer_get_status(longo target)
{
	GLenum result = glCheckFramebufferStatus(gl_fb_target(target));
	switch(result)
	{
	case GL_FRAMEBUFFER_COMPLETE: return FRAME_BUFFER_COMPLETE;
	case GL_FRAMEBUFFER_UNDEFINED: return FRAME_BUFFER_UNDEFINED;
	case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT: return FRAME_BUFFER_INCOMPLETE_ATTACHMENT;
	case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT: return FRAME_BUFFER_INCOMPLETE_MISSING_ATTACHMENT;
	case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER: return FRAME_BUFFER_INCOMPLETE_DRAW_BUFFER;
	case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER: return FRAME_BUFFER_INCOMPLETE_READ_BUFFER;
	case GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE: return FRAME_BUFFER_INCOMPLETE_MULTISAMPLE;
	case GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS: return FRAME_BUFFER_INCOMPLETE_LAYER_TARGETS;
	case GL_FRAMEBUFFER_UNSUPPORTED: return FRAME_BUFFER_UNSUPPORTED;
	}
	return FRAME_BUFFER_COMPLETE;
}
void frame_buffer_blit(longo src_x, longo src_y, longo src_w, longo src_h, longo dst_x, longo dst_y, longo dst_w, longo dst_h, longo mask, longo filter)
{
	glBlitFramebuffer(src_x, src_y, src_x+src_w, src_y+src_h, dst_x, dst_y, dst_x+dst_w, dst_y+dst_h, gl_buffer_flags(mask), gl_filter(filter));
}
//render_buffer
quadro render_buffer_create() { quadro id; glGenRenderbuffers(1, &id); return id; }
void render_buffer_destroy(quadro id) { glDeleteRenderbuffers(1, &id); }
void render_buffer_bind(quadro id) { glBindRenderbuffer(GL_RENDERBUFFER, id); }
void render_buffer_unbind() { glBindRenderbuffer(GL_RENDERBUFFER, 0); }
void render_buffer_storage(longo format, longo width, longo height)
{
	glRenderbufferStorage(GL_RENDERBUFFER, gl_format(format), width, height);
}
void render_buffer_storage_multisample(longo format, longo width, longo height, longo samples)
{
	glRenderbufferStorageMultisample(GL_RENDERBUFFER, samples, gl_format(format), width, height);
}
//shader
#define STAGE_VERTEX 0
#define STAGE_FRAGMENT 1
#define STAGE_GEOMETRY 2
#define FIND_STAGE \
	if(current == STAGE_VERTEX) \
		vertex_length = (i - vertex_offset); \
	if(current == STAGE_FRAGMENT) \
		fragment_length = (i - fragment_offset); \
	if(current == STAGE_GEOMETRY) \
		geometry_length = (i - geometry_offset);
typedef struct Error {
	byte success;
	char* error;
	inta error_length;
} Error;
quadro shader_create() { return glCreateProgram(); }
static byte stage_compile(GLenum stage_type, GLuint* out_stage, 
	const char* const source, int source_length,
	char* error, int error_max_length, inta* error_length)
{
	GLuint stage = glCreateShader(stage_type);
	glShaderSource(stage, 1, &source, &source_length);
	glCompileShader(stage);
	GLint status = 0;
	glGetShaderiv(stage, GL_COMPILE_STATUS, &status);
	if(status == 0)
	{
		char* tmp;
		if(stage_type == GL_VERTEX_SHADER) tmp = "failed to compile vertex stage\n";
		else if(stage_type == GL_FRAGMENT_SHADER) tmp = "failed to compile fragment stage\n";
		else if(stage_type == GL_GEOMETRY_SHADER) tmp = "failed to compile geometry stage\n";
		int tmp_len = strlen(tmp);
		for(int k = 0; k < tmp_len; k++)
		{
			error[*error_length + k] = tmp[k];
		}
		(*error_length) += tmp_len;
		GLsizei actual_length = 0;
		glGetShaderInfoLog(stage, error_max_length, &actual_length, &error[*error_length]);
		(*error_length) += actual_length;
		glDeleteShader(stage);
		return 0;
	}
	*out_stage = stage;
	return 1;
}
void shader_source(octa* r, quadro id, octa* source_buf, octa* error_buf)
{
	char* source = (char*)source_buf[0];
	int source_length = (int)source_buf[1];
	char* error = (char*)error_buf[0];
	int error_max_length = (int)error_buf[1];
	//find the vertex, fragment and geometry stages
	char* vertex;
	char* fragment;
	char* geometry;
	int current=-1;
	int vertex_offset=0, fragment_offset=0, geometry_offset=0;
	int vertex_length=0, fragment_length=0, geometry_length=0;
	int i;
	for(i = 0; i < source_length; i++)
	{
		if(source[i] == '#' && i+8 < source_length
			&& source[i+1] == 'v' && source[i+2] == 'e' && source[i+3] == 'r'
			&& source[i+4] == 't' && source[i+5] == 'e' && source[i+6] == 'x'
			&& (source[i+7] == '\n' || source[i+7] == '\r' || source[i+7] == '\t' || source[i+7] == ' '))
		{
			FIND_STAGE
			current = STAGE_VERTEX;
			vertex_offset = i+8;
			vertex = &source[vertex_offset];
			i+=7;
		}
		if(source[i] == '#' && i+10 < source_length
			&& source[i+1] == 'f' && source[i+2] == 'r' && source[i+3] == 'a'
			&& source[i+4] == 'g' && source[i+5] == 'm' && source[i+6] == 'e'
			&& source[i+7] == 'n' && source[i+8] == 't'
			&& (source[i+9] == '\n' || source[i+9] == '\r' || source[i+9] == '\t' || source[i+9] == ' '))
		{
			FIND_STAGE
			current = STAGE_FRAGMENT;
			fragment_offset = i+10;
			fragment = &source[fragment_offset];
			i+=9;
		}
		if(source[i] == '#' && i+10 < source_length
			&& source[i+1] == 'g' && source[i+2] == 'e' && source[i+3] == 'o'
			&& source[i+4] == 'm' && source[i+5] == 'e' && source[i+6] == 't'
			&& source[i+7] == 'r' && source[i+8] == 'y'
			&& (source[i+9] == '\n' || source[i+9] == '\r' || source[i+9] == '\t' || source[i+9] == ' '))
		{
			FIND_STAGE
			current = STAGE_GEOMETRY;
			geometry_offset = i+10;
			geometry = &source[geometry_offset];
			i+=9;
		}
	}
	FIND_STAGE
	//create the stages, compile & link
	Error result;
	result.success = 1;
	result.error = 0;
	result.error_length = 0;
	inta error_length = 0;
	GLuint vs = 0, fs = 0, gs = 0;
	if(vertex_length > 0)
		result.success &= stage_compile(GL_VERTEX_SHADER, &vs, vertex, vertex_length, 
			error, error_max_length, &error_length);
	if(fragment_length > 0)
		result.success &= stage_compile(GL_FRAGMENT_SHADER, &fs, fragment, fragment_length, 
			error, error_max_length, &error_length);
	if(geometry_length > 0)
		result.success &= stage_compile(GL_GEOMETRY_SHADER, &gs, geometry, geometry_length, 
			error, error_max_length, &error_length);
	if(vs != 0) glAttachShader(id, vs);
	if(fs != 0) glAttachShader(id, fs);
	if(gs != 0) glAttachShader(id, gs);
	glLinkProgram(id);
	GLint status = 0;
	glGetProgramiv(id, GL_LINK_STATUS, &status);
	if(status == 0)
	{
		GLsizei actual_length = 0;
		glGetProgramInfoLog(id, error_max_length, &actual_length, &error[error_length]);
		error_length += actual_length;
		result.success = 0;
	}
	if(vs != 0) { glDetachShader(id, vs); glDeleteShader(vs); }
	if(fs != 0) { glDetachShader(id, fs); glDeleteShader(fs); }
	if(gs != 0) { glDetachShader(id, gs); glDeleteShader(gs); }
	if(!result.success)
	{
		result.error = error;
		result.error_length = error_length;
	}
	memcpy(r, &result, sizeof(Error));
}
void shader_destroy(quadro id) { glDeleteProgram(id); }
void shader_bind(quadro id) { glUseProgram(id); }
void shader_unbind() { glUseProgram(0); }
longo shader_get_attribute_location(quadro id, octa* name_buf)
{
	String name;
	memcpy(&name, name_buf, 16);
	return glGetAttribLocation(id, to_cstr(name));
}
longo shader_get_output_location(quadro id, octa* name_buf)
{
	String name;
	memcpy(&name, name_buf, 16);
	return glGetFragDataLocation(id, to_cstr(name));
}
longo shader_get_uniform_location(quadro id, octa* name_buf)
{
	String name;
	memcpy(&name, name_buf, 16);
	return glGetUniformLocation(id, to_cstr(name));
}
longo shader_get_uniform_block_index(quadro id, octa* name_buf)
{
	String name;
	memcpy(&name, name_buf, 16);
	return glGetUniformBlockIndex(id, to_cstr(name));
}
void shader_uniform_block_binding(quadro id, longo index, longo binding)
{
	glUniformBlockBinding(id, index, binding);
}
void shader_uniform1f(longo location, quadro x)
{
	glUniform1f(location, *(float*)&x);
}
void shader_uniform2f(longo location, quadro x, quadro y)
{
	glUniform2f(location, *(float*)&x, *(float*)&y);
}
void shader_uniform3f(longo location, quadro x, quadro y, quadro z)
{
	glUniform3f(location, *(float*)&x, *(float*)&y, *(float*)&z);
}
void shader_uniform4f(longo location, quadro x, quadro y, quadro z, quadro w)
{
	glUniform4f(location, *(float*)&x, *(float*)&y, *(float*)&z, *(float*)&w);
}
void shader_uniform1fa(longo location, longo num, octa* values)
{
	glUniform1fv(location, num, (float*)values[0]);
}
void shader_uniform2fa(longo location, longo num, octa* values)
{
	glUniform2fv(location, num, (float*)values[0]);
}
void shader_uniform3fa(longo location, longo num, octa* values)
{
	glUniform3fv(location, num, (float*)values[0]);
}
void shader_uniform4fa(longo location, longo num, octa* values)
{
	glUniform4fv(location, num, (float*)values[0]);
}
void shader_uniform1i(longo location, longo x)
{
	glUniform1i(location, x);
}
void shader_uniform2i(longo location, longo x, longo y)
{
	glUniform2i(location, x, y);
}
void shader_uniform3i(longo location, longo x, longo y, longo z)
{
	glUniform3i(location, x, y, z);
}
void shader_uniform4i(longo location, longo x, longo y, longo z, longo w)
{
	glUniform4i(location, x, y, z, w);
}
void shader_uniform1ia(longo location, longo num, octa* values)
{
	glUniform1iv(location, num, (longo*)values[0]);
}
void shader_uniform2ia(longo location, longo num, octa* values)
{
	glUniform2iv(location, num, (longo*)values[0]);
}
void shader_uniform3ia(longo location, longo num, octa* values)
{
	glUniform3iv(location, num, (longo*)values[0]);
}
void shader_uniform4ia(longo location, longo num, octa* values)
{
	glUniform4iv(location, num, (longo*)values[0]);
}
void shader_uniform1q(longo location, quadro x)
{
	glUniform1ui(location, x);
}
void shader_uniform2q(longo location, quadro x, quadro y)
{
	glUniform2ui(location, x, y);
}
void shader_uniform3q(longo location, quadro x, quadro y, quadro z)
{
	glUniform3ui(location, x, y, z);
}
void shader_uniform4q(longo location, quadro x, quadro y, quadro z, quadro w)
{
	glUniform4ui(location, x, y, z, w);
}
void shader_uniform1qa(longo location, longo num, octa* values)
{
	glUniform1uiv(location, num, (quadro*)values[0]);
}
void shader_uniform2qa(longo location, longo num, octa* values)
{
	glUniform2uiv(location, num, (quadro*)values[0]);
}
void shader_uniform3qa(longo location, longo num, octa* values)
{
	glUniform3uiv(location, num, (quadro*)values[0]);
}
void shader_uniform4qa(longo location, longo num, octa* values)
{
	glUniform4uiv(location, num, (quadro*)values[0]);
}
void shader_uniform_matrix2fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix2fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix3fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix3fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix4fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix4fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix2x3fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix2x3fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix3x2fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix3x2fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix2x4fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix2x4fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix4x2fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix4x2fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix3x4fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix3x4fv(location, num, transpose, (float*)values[0]);
}
void shader_uniform_matrix4x3fa(longo location, longo num, byte transpose, octa* values)
{
	glUniformMatrix4x3fv(location, num, transpose, (float*)values[0]);
}
//buffers
static longo gl_buffer_target(longo target)
{
	switch(target)
	{
	case BUFFER_ARRAY: return GL_ARRAY_BUFFER;
	case BUFFER_ELEMENT_ARRAY: return GL_ELEMENT_ARRAY_BUFFER;
	case BUFFER_COPY_WRITE: return GL_COPY_WRITE_BUFFER;
	case BUFFER_COPY_READ: return GL_COPY_READ_BUFFER;
	case BUFFER_PIXEL_UNPACK: return GL_PIXEL_UNPACK_BUFFER;
	case BUFFER_PIXEL_PACK: return GL_PIXEL_PACK_BUFFER;
	case BUFFER_TEXTURE: return GL_TEXTURE_BUFFER;
	case BUFFER_UNIFORM: return GL_UNIFORM_BUFFER;
	}
	return 0;
}
quadro buffer_create() { quadro id; glGenBuffers(1, &id); return id; }
void buffer_destroy(quadro id) { glDeleteBuffers(1, &id); }
void buffer_bind(longo target, quadro id)
{
	glBindBuffer(gl_buffer_target(target), id);
}
void buffer_unbind(longo target)
{
	glBindBuffer(gl_buffer_target(target), 0);
}
void buffer_bind_base(longo target, longo index, quadro id)
{
	glBindBufferBase(gl_buffer_target(target), index, id);
}
void buffer_bind_range(longo target, longo index, quadro id, inta offset, inta size)
{
	glBindBufferRange(gl_buffer_target(target), index, id, offset, size);
}
static longo gl_usage(longo usage)
{
	switch(usage)
	{
	case USAGE_STATIC_DRAW: return GL_STATIC_DRAW;
	case USAGE_STATIC_READ: return GL_STATIC_READ;
	case USAGE_STATIC_COPY: return GL_STATIC_COPY;
	case USAGE_DYNAMIC_DRAW: return GL_DYNAMIC_DRAW;
	case USAGE_DYNAMIC_READ: return GL_DYNAMIC_READ;
	case USAGE_DYNAMIC_COPY: return GL_DYNAMIC_COPY;
	case USAGE_STREAM_DRAW: return GL_STREAM_DRAW;
	case USAGE_STREAM_READ: return GL_STREAM_READ;
	case USAGE_STREAM_COPY: return GL_STREAM_COPY;
	}
	return 0;
}
void buffer_data(longo target, inta size, octa data, longo usage)
{
	glBufferData(gl_buffer_target(target), size, (void*)data, gl_usage(usage));
}
void buffer_sub_data(longo target, inta offset, inta size, octa data)
{
	glBufferSubData(gl_buffer_target(target), offset, size, (void*)data);
}
static longo gl_access(longo access)
{
	switch(access)
	{
	case ACCESS_WRITE_ONLY: return GL_WRITE_ONLY;
	case ACCESS_READ_ONLY: return GL_READ_ONLY;
	case ACCESS_WRITE_READ: return GL_READ_WRITE;
	}
	return 0;
}
void buffer_map(octa* r, longo target, inta size, longo access)
{
	octa region[2];
	region[0] = (octa)glMapBuffer(gl_buffer_target(target), gl_access(access));
	region[1] = size;
	memcpy(r, region, 16);
}
byte buffer_unmap(longo target)
{
	return glUnmapBuffer(gl_buffer_target(target));
}
void buffer_copy_sub_data(
	longo read_target, longo write_target,
	inta read_offset, inta write_offset, inta size)
{
	glCopyBufferSubData(
		gl_buffer_target(read_target), gl_buffer_target(write_target),
		read_offset, write_offset, size);
}
//vertex arrays
quadro vertex_array_create() { quadro id; glGenVertexArrays(1, &id); return id; }
void vertex_array_destroy(quadro id) { glDeleteVertexArrays(1, &id); }
void vertex_array_bind(quadro id) { glBindVertexArray(id); }
void vertex_array_unbind() { glBindVertexArray(0); }
//pipeline
void pipeline_clear(longo buffer_flags)
{
	glClear(gl_buffer_flags(buffer_flags));
}
void pipeline_clear_color(quadro r, quadro g, quadro b, quadro a)
{
	glClearColor(*(float*)&r, *(float*)&g, *(float*)&b, *(float*)&a);
}
void pipeline_clear_stencil(byte value)
{
	glClearStencil(value);
}
void pipeline_clear_depth(quadro depth)
{
	glClearDepth(*(float*)&depth);
}
void pipeline_depth_enable(byte enable)
{
	if(enable)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);
}
static longo gl_func(longo func)
{
	switch(func)
	{
	case TEST_NEVER: return GL_NEVER;
	case TEST_ALWAYS: return GL_ALWAYS;
	case TEST_LESS: return GL_LESS;
	case TEST_LESS_OR_EQUAL: return GL_LEQUAL;
	case TEST_GREATER: return GL_GREATER;
	case TEST_GREATER_OR_EQUAL: return GL_GEQUAL;
	case TEST_EQUAL: return GL_EQUAL;
	case TEST_NOT_EQUAL: return GL_NOTEQUAL;
	}
	return 0;
}
void pipeline_depth_func(longo func)
{
	glDepthFunc(gl_func(func));
}
void pipeline_depth_mask(byte mask)
{
	glDepthMask(mask);
}
void pipeline_depth_clamp(byte clamp)
{
	if(clamp)
		glEnable(GL_DEPTH_CLAMP);
	else
		glDisable(GL_DEPTH_CLAMP);
}
void pipeline_depth_range(quadro near_val, quadro far_val)
{
	glDepthRange(*(float*)&near_val, *(float*)&far_val);
}
void pipeline_stencil_enable(byte enable)
{
	if(enable)
		glEnable(GL_STENCIL_TEST);
	else
		glDisable(GL_STENCIL_TEST);
}
static longo gl_face(longo face)
{
	switch(face)
	{
	case FACE_FRONT: return GL_FRONT;
	case FACE_BACK: return GL_BACK;
	case FACE_FRONT_BACK: return GL_FRONT_AND_BACK;
	}
	return 0;
}
static longo gl_stencil(longo stencil)
{
	switch(stencil)
	{
	case STENCIL_KEEP: return GL_KEEP;
	case STENCIL_ZERO: return GL_ZERO;
	case STENCIL_REPLACE: return GL_REPLACE;
	case STENCIL_INCR: return GL_INCR;
	case STENCIL_INCR_WRAP: return GL_INCR_WRAP;
	case STENCIL_DECR: return GL_DECR;
	case STENCIL_DECR_WRAP: return GL_DECR_WRAP;
	case STENCIL_INVERT: return GL_INVERT;
	}
	return 0;
}
void pipeline_stencil_func(longo func, byte ref, byte mask)
{
	glStencilFunc(gl_func(func), ref, mask);
}
void pipeline_stencil_op(longo sfail, longo dpfail, longo dppass)
{
	glStencilOp(gl_stencil(sfail), gl_stencil(dpfail), gl_stencil(dppass));
}
void pipeline_stencil_mask(byte mask)
{
	glStencilMask(mask);
}
void pipeline_stencil_func_separate(longo face, longo func, byte ref, byte mask)
{
	glStencilFuncSeparate(gl_face(face), gl_func(func), ref, mask);
}
void pipeline_stencil_op_separate(longo face, longo sfail, longo dpfail, longo dppass)
{
	glStencilOpSeparate(gl_face(face), gl_stencil(sfail), gl_stencil(dpfail), gl_stencil(dppass));
}
void pipeline_stencil_mask_separate(longo face, byte mask)
{
	glStencilMaskSeparate(gl_face(face), mask);
}
void pipeline_blend_enable(byte enable)
{
	if(enable)
		glEnable(GL_BLEND);
	else
		glDisable(GL_BLEND);
}
static longo gl_blend_factor(longo factor)
{
	switch(factor)
	{
	case BLEND_FACTOR_ZERO: return GL_ZERO;
	case BLEND_FACTOR_ONE: return GL_ONE;
	case BLEND_FACTOR_SRC_COLOR: return GL_SRC_COLOR;
	case BLEND_FACTOR_SRC_ALPHA: return GL_SRC_ALPHA;
	case BLEND_FACTOR_DST_COLOR: return GL_DST_COLOR;
	case BLEND_FACTOR_DST_ALPHA: return GL_DST_ALPHA;
	case BLEND_FACTOR_CONSTANT_COLOR: return GL_CONSTANT_COLOR;
	case BLEND_FACTOR_CONSTANT_ALPHA: return GL_CONSTANT_ALPHA;
	case BLEND_FACTOR_INV_SRC_COLOR: return GL_ONE_MINUS_SRC_COLOR;
	case BLEND_FACTOR_INV_SRC_ALPHA: return GL_ONE_MINUS_SRC_ALPHA;
	case BLEND_FACTOR_INV_DST_COLOR: return GL_ONE_MINUS_DST_COLOR;
	case BLEND_FACTOR_INV_DST_ALPHA: return GL_ONE_MINUS_DST_ALPHA;
	case BLEND_FACTOR_INV_CONSTANT_COLOR: return GL_ONE_MINUS_CONSTANT_COLOR;
	case BLEND_FACTOR_INV_CONSTANT_ALPHA: return GL_ONE_MINUS_CONSTANT_ALPHA;
	}
	return 0;
}
static longo gl_blend_equation(longo equation)
{
	switch(equation)
	{
	case BLEND_EQUATION_ADD: return GL_FUNC_ADD;
	case BLEND_EQUATION_SUBTRACT: return GL_FUNC_SUBTRACT;
	case BLEND_EQUATION_REVERSE_SUBTRACT: return GL_FUNC_REVERSE_SUBTRACT;
	case BLEND_EQUATION_MIN: return GL_MIN;
	case BLEND_EQUATION_MAX: return GL_MAX;
	}
	return 0;
}
void pipeline_blend_factor(longo src_factor, longo dst_factor)
{
	glBlendFunc(gl_blend_factor(src_factor), gl_blend_factor(dst_factor));
}
void pipeline_blend_equation(longo equation)
{
	glBlendEquation(gl_blend_equation(equation));
}
void pipeline_blend_factor_separate(
	longo src_rgb, longo dst_rgb,
	longo src_alpha, longo dst_alpha)
{
	glBlendFuncSeparate(
		gl_blend_factor(src_rgb), gl_blend_factor(dst_rgb),
		gl_blend_factor(src_alpha), gl_blend_factor(dst_alpha));
}
void pipeline_blend_equation_separate(longo equation_rgb, longo equation_alpha)
{
	glBlendEquationSeparate(gl_blend_equation(equation_rgb), gl_blend_equation(equation_alpha));
}
void pipeline_blend_color(quadro r, quadro g, quadro b, quadro a)
{
	glBlendColor(*(float*)&r, *(float*)&g, *(float*)&b, *(float*)&a);
}
void pipeline_color_clamp(byte clamp)
{
	glClampColor(GL_CLAMP_READ_COLOR, clamp);
}
void pipeline_color_mask(byte r, byte g, byte b, byte a)
{
	glColorMask(r, g, b, a);
}
void pipeline_logic_enable(byte enable)
{
	if(enable)
		glEnable(GL_COLOR_LOGIC_OP);
	else
		glDisable(GL_COLOR_LOGIC_OP);
}
static longo gl_logic(longo op)
{
	switch(op)
	{
	case LOGIC_CLEAR: return GL_CLEAR;
	case LOGIC_SET: return GL_SET;
	case LOGIC_COPY: return GL_COPY;
	case LOGIC_COPY_INVERTED: return GL_COPY_INVERTED;
	case LOGIC_NOOP: return GL_NOOP;
	case LOGIC_INVERT: return GL_INVERT;
	case LOGIC_AND: return GL_AND;
	case LOGIC_NAND: return GL_NAND;
	case LOGIC_OR: return GL_OR;
	case LOGIC_NOR: return GL_NOR;
	case LOGIC_XOR: return GL_XOR;
	case LOGIC_EQUIV: return GL_EQUIV;
	case LOGIC_AND_REVERSE: return GL_AND_REVERSE;
	case LOGIC_AND_INVERTED: return GL_AND_INVERTED;
	case LOGIC_OR_REVERSE: return GL_OR_REVERSE;
	case LOGIC_OR_INVERTED: return GL_OR_INVERTED;
	}
	return 0;
}
void pipeline_logic_op(longo op)
{
	glLogicOp(gl_logic(op));
}
void pipeline_cull_enable(byte enable)
{
	if(enable)
		glEnable(GL_CULL_FACE);
	else
		glDisable(GL_CULL_FACE);
}
void pipeline_cull_face(longo face)
{
	glCullFace(gl_face(face));
} 
void pipeline_cull_orientation(byte ccw)
{
	glFrontFace(ccw ? GL_CCW : GL_CW);
}
static longo gl_polygon(longo mode)
{
	switch(mode)
	{
	case POLYGON_FILL: return GL_FILL;
	case POLYGON_WIREFRAME: return GL_LINE;
	case POLYGON_POINT: return GL_POINT;
	}
	return 0;
}
void pipeline_polygon_mode(longo face, longo mode)
{
	glPolygonMode(gl_face(face), gl_polygon(mode));
}
void pipeline_viewport(longo x, longo y, longo width, longo height)
{
	glViewport(x, y, width, height);
}
void pipeline_scissor_enable(byte enable)
{
	if(enable)
		glEnable(GL_SCISSOR_TEST);
	else
		glDisable(GL_SCISSOR_TEST);
}
void pipeline_scissor_rect(longo x, longo y, longo width, longo height)
{
	glScissor(x, y, width, height);
}
void pipeline_point_shader_enable(byte enable)
{
	if(enable)
		glEnable(GL_PROGRAM_POINT_SIZE);
	else
		glDisable(GL_PROGRAM_POINT_SIZE);
}
void pipeline_point_size(quadro size)
{
	glPointSize(*(float*)&size);
}
void pipeline_point_coord_origin(longo origin)
{
	glPointParameteri(GL_POINT_SPRITE_COORD_ORIGIN,
		origin != 0 ? GL_LOWER_LEFT : GL_UPPER_LEFT);
}
void pipeline_point_fade_threshold_size(quadro size)
{
	glPointParameterf(GL_POINT_FADE_THRESHOLD_SIZE, *(float*)&size);
}
void pipeline_primitive_restart_enable(byte enable)
{
	if(enable)
		glEnable(GL_PRIMITIVE_RESTART);
	else
		glDisable(GL_PRIMITIVE_RESTART);
}
void pipeline_primitive_restart_index(longo index)
{
	glPrimitiveRestartIndex(index);
}
void pipeline_clip_plane_enable(longo index, byte enable)
{
	if(enable)
		glEnable(GL_CLIP_DISTANCE0 + (index & 0x3fff));
	else
		glDisable(GL_CLIP_DISTANCE0 + (index & 0x3fff));
}
void pipeline_multisample_enable(byte enable)
{
	if(enable)
		glEnable(GL_MULTISAMPLE);
	else
		glDisable(GL_MULTISAMPLE);
}
void pipeline_present()
{
	SDL_GL_SwapWindow(sdl_window);
}
void pipeline_finish()
{
	glFinish();
}
static longo gl_pixel_buffer(longo buffer)
{
	if(buffer < 0)
		return GL_NONE;
	else
		return GL_COLOR_ATTACHMENT0 + buffer;
}
static const GLenum* gl_draw_buffers(longo* buffers, tiny num)
{
	GLenum* res = (GLenum*)&temporary_buffer[0];
	for(tiny i = 0; i < num; i++)
	{
		res[i] = gl_pixel_buffer(buffers[i]);
	}
	return res;
}
void pipeline_draw_buffers(octa* buffers)
{
	glDrawBuffers(buffers[1], gl_draw_buffers((longo*)buffers[0], buffers[1]));
}
void pipeline_read_buffer(longo buffer_index)
{
	glReadBuffer(gl_pixel_buffer(buffer_index));
}
static longo gl_pixel_parameter(longo parameter)
{
	switch(parameter)
	{
	case PIXEL_PACK_SWAP_BYTES: return GL_PACK_SWAP_BYTES;
	case PIXEL_PACK_LSB_FIRST: return GL_PACK_LSB_FIRST;
	case PIXEL_PACK_ROW_LENGTH: return GL_PACK_ROW_LENGTH;
	case PIXEL_PACK_IMAGE_HEIGHT: return GL_PACK_IMAGE_HEIGHT;
	case PIXEL_PACK_SKIP_PIXELS: return GL_PACK_SKIP_PIXELS;
	case PIXEL_PACK_SKIP_ROWS: return GL_PACK_SKIP_ROWS;
	case PIXEL_PACK_SKIP_IMAGES: return GL_PACK_SKIP_IMAGES;
	case PIXEL_PACK_ALIGNMENT: return GL_PACK_ALIGNMENT;
	case PIXEL_UNPACK_SWAP_BYTES: return GL_UNPACK_SWAP_BYTES;
	case PIXEL_UNPACK_LSB_FIRST: return GL_UNPACK_LSB_FIRST;
	case PIXEL_UNPACK_ROW_LENGTH: return GL_UNPACK_ROW_LENGTH;
	case PIXEL_UNPACK_IMAGE_HEIGHT: return GL_UNPACK_IMAGE_HEIGHT;
	case PIXEL_UNPACK_SKIP_PIXELS: return GL_UNPACK_SKIP_PIXELS;
	case PIXEL_UNPACK_SKIP_ROWS: return GL_UNPACK_SKIP_ROWS;
	case PIXEL_UNPACK_SKIP_IMAGES: return GL_UNPACK_SKIP_IMAGES;
	case PIXEL_UNPACK_ALIGNMENT: return GL_UNPACK_ALIGNMENT;
	}
	return 0;
}
void pipeline_pixel_store(longo parameter, longo value)
{
	glPixelStorei(gl_pixel_parameter(parameter), value);
}
void pipeline_read_pixels(longo x, longo y, longo width, longo height, longo components, longo data_type, octa data)
{
	glReadPixels(x, y, width, height, gl_components(components), gl_data_type(data_type), (void*)data);
}
tiny pipeline_get_error()
{
	GLenum err = glGetError();
	if(err != 0)
		return err-0x4ff;
	return 0;
}
//attributes
void attribute_enable(longo index, byte enable)
{
	if(enable)
		glEnableVertexAttribArray(index);
	else
		glDisableVertexAttribArray(index);
}
void attribute_format(
	longo index, longo num_components, longo data_type,
	byte normalized, longo stride, longo pos)
{
	inta ipos = pos;
	glVertexAttribPointer(index, num_components, gl_data_type(data_type), normalized, stride, (void*)ipos);
}
void attribute_integer_format(
	longo index, longo num_components, longo data_type, longo stride, longo pos)
{
	inta ipos = pos;
	glVertexAttribIPointer(index, num_components, gl_data_type(data_type), stride, (void*)ipos);
}
void attribute_divisor(longo index, longo divisor)
{
	glVertexAttribDivisor(index, divisor);
}
//drawing commands
#define gl_primitive(x) x
void draw_arrays(longo primitive, longo start, longo num_indices, longo num_instances)
{
	if(num_instances == 1)
		glDrawArrays(gl_primitive(primitive), start, num_indices);
	else
		glDrawArraysInstanced(gl_primitive(primitive), start, num_indices, num_instances);
}
#define ELEMENTS_ARGS gl_primitive(primitive), num_indices, gl_data_type(data_type), (const void*)istart
void draw_elements(longo primitive, longo data_type, longo start, longo num_indices, longo base_vertex, longo num_instances)
{
	inta istart = start;
	if(base_vertex == 0)
	{
		if(num_instances == 1)
			glDrawElements(ELEMENTS_ARGS);
		else
			glDrawElementsInstanced(ELEMENTS_ARGS, num_instances);
	}
	else
	{
		if(num_instances == 1)
			glDrawElementsBaseVertex(ELEMENTS_ARGS, base_vertex);
		else
			glDrawElementsInstancedBaseVertex(ELEMENTS_ARGS, num_instances, base_vertex);
	}
}
static void init_graphics_resources()
{
	glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS);
}
//audio device
static SDL_AudioFormat to_sdl_audio_format(byte format)
{
	switch(format)
	{
	case AUDIO_FORMAT_BYTE: return AUDIO_U8;
	case AUDIO_FORMAT_DOUBLE: return AUDIO_U16;
	case AUDIO_FORMAT_QUADRO: return AUDIO_S32; //AUDIO_U32 does not exists in SDL
	case AUDIO_FORMAT_TINY: return AUDIO_S8;
	case AUDIO_FORMAT_SHORT: return AUDIO_S16;
	case AUDIO_FORMAT_LONG: return AUDIO_S32;
	case AUDIO_FORMAT_FLOAT: return AUDIO_F32;
	}
	return AUDIO_S16;
}
SDL_AudioDeviceID audio_player_id;
SDL_AudioDeviceID audio_recorder_id;
byte audio_open(byte format, longo num_channels, longo sample_rate, longo num_samples, void (*callback)(void*,byte*,int), octa userdata)
{
	if(format == AUDIO_FORMAT_QUADRO || format > AUDIO_FORMAT_FLOAT)
	{
		//invalid format
		return false;
	}
	SDL_AudioSpec desired;
	memset(&desired, 0, sizeof(SDL_AudioSpec));
	desired.freq = sample_rate;
	desired.format = to_sdl_audio_format(format);
	desired.channels = num_channels;
	desired.samples = num_samples;
	desired.callback = callback;
	desired.userdata = (void*)userdata;
	audio_player_id = SDL_OpenAudioDevice(NULL, 0, &desired, NULL, 0);
	return audio_player_id != 0;
}
void audio_play()
{
	SDL_PauseAudioDevice(audio_player_id, false);
}
void audio_pause()
{
	SDL_PauseAudioDevice(audio_player_id, true);
}
void audio_close()
{
	SDL_CloseAudioDevice(audio_player_id);
}
void audio_lock()
{
	SDL_LockAudioDevice(audio_player_id);
}
void audio_unlock()
{
	SDL_UnlockAudioDevice(audio_player_id);
}
byte audio_recorder_open(byte format, longo num_channels, longo sample_rate, longo num_samples, void (*callback)(void*,byte*,int), octa userdata)
{
	SDL_AudioSpec desired;
	memset(&desired, 0, sizeof(SDL_AudioSpec));
	desired.freq = sample_rate;
	desired.format = to_sdl_audio_format(format);
	desired.channels = num_channels;
	desired.samples = num_samples;
	desired.callback = callback;
	desired.userdata = (void*)userdata;
	audio_recorder_id = SDL_OpenAudioDevice(NULL, true, &desired, NULL, 0);
	return audio_recorder_id != 0;
}
void audio_recorder_record()
{
	SDL_PauseAudioDevice(audio_recorder_id, false);
}
void audio_recorder_pause()
{
	SDL_PauseAudioDevice(audio_recorder_id, true);
}
void audio_recorder_close()
{
	SDL_CloseAudioDevice(audio_recorder_id);
}
void audio_recorder_lock()
{
	SDL_LockAudioDevice(audio_recorder_id);
}
void audio_recorder_unlock()
{
	SDL_UnlockAudioDevice(audio_recorder_id);
}
//threading
octa thread_create(longo (*thread_proc)(void*), octa data)
{
	return (octa)SDL_CreateThread(thread_proc, "thread", (void*)data);
}
void thread_detach(octa thread)
{
	SDL_DetachThread((SDL_Thread*)thread);
}
void thread_wait(octa thread, longo* return_value)
{
	SDL_WaitThread((SDL_Thread*)thread, return_value);
}
void thread_set_priority(byte priority)
{
	Uint32 sdl_thread_priority = 0;
	if(priority == ENUMS_THREAD_PRIORITY_LOW)
		sdl_thread_priority = SDL_THREAD_PRIORITY_LOW;
	if(priority == ENUMS_THREAD_PRIORITY_NORMAL)
		sdl_thread_priority = SDL_THREAD_PRIORITY_NORMAL;
	if(priority == ENUMS_THREAD_PRIORITY_HIGH)
		sdl_thread_priority = SDL_THREAD_PRIORITY_HIGH;
	SDL_SetThreadPriority(sdl_thread_priority);
}
longo thread_current_id() { return SDL_ThreadID(); }
longo thread_id(octa thread) { return SDL_GetThreadID((SDL_Thread*)thread); }
octa cond_create() { return (octa)SDL_CreateCond(); }
void cond_destroy(octa cond) { SDL_DestroyCond((SDL_cond*)cond); }
void cond_broadcast(octa cond) { SDL_CondBroadcast((SDL_cond*)cond); }
void cond_signal(octa cond) { SDL_CondSignal((SDL_cond*)cond); }
void cond_wait(octa cond, octa m) { SDL_CondWait((SDL_cond*)cond, (SDL_mutex*)m); }
byte cond_wait_timeout(octa cond, octa m, inta timeout) { return SDL_CondWaitTimeout((SDL_cond*)cond, (SDL_mutex*)m, timeout) == 0; }
octa mutex_create() { return (octa)SDL_CreateMutex(); }
void mutex_destroy(octa m) { SDL_DestroyMutex((SDL_mutex*)m); }
void mutex_lock(octa m) { SDL_LockMutex((SDL_mutex*)m); }
byte mutex_trylock(octa m) { return SDL_TryLockMutex((SDL_mutex*)m) == 0; }
void mutex_unlock(octa m) { SDL_UnlockMutex((SDL_mutex*)m); }
octa semaphore_create(quadro val) { return (octa)SDL_CreateSemaphore(val); }
void semaphore_destroy(octa s) { SDL_DestroySemaphore((SDL_sem*)s); }
void semaphore_post(octa s) { SDL_SemPost((SDL_sem*)s); }
byte semaphore_trywait(octa s) { return SDL_SemTryWait((SDL_sem*)s) == 0; }
void semaphore_wait(octa s) { SDL_SemWait((SDL_sem*)s); }
byte semaphore_wait_timeout(octa s, inta timeout) { return SDL_SemWaitTimeout((SDL_sem*)s, timeout); }
quadro semaphore_value(octa s) { return SDL_SemValue((SDL_sem*)s); }
longo atomic_set(longo* atomic, longo value) { return SDL_AtomicSet((SDL_atomic_t*)atomic, value); }
octa atomic_set_ptr(octa* atomic, octa value) { return (octa)SDL_AtomicSetPtr((void**)atomic, (void*)value); }
longo atomic_get(longo* atomic) { return SDL_AtomicGet((SDL_atomic_t*)atomic); }
octa atomic_get_ptr(octa* atomic) { return (octa)SDL_AtomicGetPtr((void**)atomic); }
longo atomic_add(longo* atomic, longo value) { return SDL_AtomicAdd((SDL_atomic_t*)atomic, value); }
void atomic_inc_ref(longo* atomic) { SDL_AtomicIncRef((SDL_atomic_t*)atomic); }
byte atomic_dec_ref(longo* atomic) { return SDL_AtomicDecRef((SDL_atomic_t*)atomic); }
byte atomic_cas(longo* atomic, longo oldvalue, longo newvalue) { return SDL_AtomicCAS((SDL_atomic_t*)atomic, oldvalue, newvalue); }
byte atomic_cas_ptr(octa* atomic, octa oldvalue, octa newvalue) { return SDL_AtomicCASPtr((void**)atomic, (void*)oldvalue, (void*)newvalue); }
void atomic_lock(longo* atomic) { SDL_AtomicLock(/*(SDL_SpinLock*)*/atomic); }
byte atomic_trylock(longo* atomic) { return SDL_AtomicTryLock(/*(SDL_SpinLock*)*/atomic); }
void atomic_unlock(longo* atomic) { SDL_AtomicUnlock(/*(SDL_SpinLock*)*/atomic); }
void atomic_compiler_barrier() { SDL_CompilerBarrier(); }
//network
typedef struct Network_Address
{
	doubli family;
	byte data[30];
} Network_Address;
void network_address_find(doubli* r, octa* name, doubli port, byte flags)
{
	String s;
	memcpy(&s, name, 16);
	struct addrinfo hints, *info = NULL;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	if(flags & NETWORK_IP4)
		hints.ai_family = AF_INET;
	if(flags & NETWORK_IP6)
		hints.ai_family = AF_INET6;
	if(flags & NETWORK_TCP)
		hints.ai_socktype = SOCK_STREAM;
	if(flags & NETWORK_UDP)
		hints.ai_socktype = SOCK_DGRAM;
	const char* host = NULL;
	if(s.length > 0)
		host = to_cstr(s);
	else
		hints.ai_flags = AI_PASSIVE;
	char service[8];
	sprintf(service, "%d", (int)port);
	int status = getaddrinfo(host, service, &hints, &info);
	memset(r, 0, sizeof(Network_Address));
	if(status == 0)
		memcpy(r, info->ai_addr, info->ai_addrlen);
	if(info != NULL) freeaddrinfo(info);
}
void network_address_get_ip(octa* r, doubli* addr)
{
	char* ip = ALLOC(0x40);
	inet_ntop(addr[0], &addr[2], ip, 0x40);
	String s = make_string_c(ip);
	memcpy(r, &s, 16);
}
doubli network_address_get_port(doubli* addr)
{
	byte* port = (byte*)&addr[1];
	return (port[0] << 8) | port[1];
}
static int find_address_length(doubli* addr)
{
	if(addr[0] == AF_INET)
		return sizeof(struct sockaddr_in);
	return sizeof(struct sockaddr_in6);
}
octa network_socket_create(byte flags)
{
	int family = AF_UNSPEC;
	int socktype = SOCK_DGRAM;
	if(flags & NETWORK_IP4)
		family = AF_INET;
	if(flags & NETWORK_IP6)
		family = AF_INET6;
	if(flags & NETWORK_TCP)
		socktype = SOCK_STREAM;
	if(flags & NETWORK_UDP)
		socktype = SOCK_DGRAM;
	octa s = socket(family, socktype, 0);
	if(s == -1)
		return NETWORK_INVALID_SOCKET;
#ifdef WINDOWS_OS
	char yes=1;
#else
	int yes=1;
#endif
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes));
	return s;
}
void network_socket_destroy(octa s)
{
#ifdef LINUX_OS
	close(s);
#endif
#ifdef WINDOWS_OS
	closesocket(s);
#endif
}
byte network_socket_failed(octa s)
{
	return s == NETWORK_INVALID_SOCKET;
}
byte network_socket_bind(octa s, doubli* addr)
{
	if(bind(s, (struct sockaddr*)addr, find_address_length(addr)) == -1)
		return false;
	return true;
}
byte network_socket_set_blocking(octa s, byte blocking)
{
#ifdef WINDOWS_OS
	unsigned long non_blocking = blocking ? 0 : 1;
	return ioctlsocket(s, FIONBIO, &non_blocking) == NO_ERROR;
#endif
#ifdef LINUX_OS
	int flags = fcntl(s, F_GETFL, 0);
	if(blocking)
		return fcntl(s, F_SETFL, flags & ~O_NONBLOCK) == 0;
	else
		return fcntl(s, F_SETFL, flags | O_NONBLOCK) == 0;
#endif
}
byte network_tcp_connect(octa s, doubli* dest)
{
	return connect(s, (struct sockaddr*)dest, find_address_length(dest)) == 0;
}
void network_tcp_disconnect(octa s)
{
	shutdown(s, 2);
}
inta network_tcp_send(octa s, inta size, octa data)
{
	return send(s, (byte*)data, size, 0);
}
inta network_tcp_receive(octa s, inta size, octa data)
{
	return recv(s, (byte*)data, (inta)size, 0);
}
byte network_tcp_listen(octa s)
{
	return listen(s, SOMAXCONN) == 0;
}
octa network_tcp_accept(octa s, doubli* out_address)
{
	socklen_t addr_size = sizeof(struct sockaddr_storage);
	octa new_s = accept(s, (struct sockaddr*)out_address, &addr_size);
	return new_s;
}
inta network_udp_send(octa s, inta size, octa data, doubli* dest)
{
	return sendto(s, (byte*)data, size, 0, (struct sockaddr*)dest, sizeof(struct sockaddr_storage));
}
inta network_udp_receive(octa s, inta size, octa data, doubli* src)
{
	int from_length = sizeof(struct sockaddr_storage);
	return recvfrom(s, (byte*)data, size, 0, (struct sockaddr*)src, (socklen_t*)&from_length);
}
#define MAX_SELECTORS 0x400
#define INVALID_SELECTOR -1
byte network_selector_failed(longo sel)
{
	return sel == INVALID_SELECTOR;
}
#ifdef WINDOWS_OS
//unfortunately an inefficient solution for non-linux platforms: select.
typedef struct Network_Selector
{
	byte alive;
	SOCKET* sockets;
	longo sockets_length, sockets_capacity;
} Network_Selector;
Network_Selector selectors[MAX_SELECTORS];
longo num_selectors;
longo network_selector_create()
{
	longo h = 0;
	for(; h < num_selectors; h++)
		if(!selectors[h].alive)
			break;
	if(h >= MAX_SELECTORS)
		return INVALID_SELECTOR;
	if(h >= num_selectors)
		num_selectors = h+1;
	selectors[h].alive = true;
	return h;
}
void network_selector_destroy(longo sel)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	selectors[sel].alive = false;
}
void network_selector_add(longo sel, octa s)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	Network_Selector* selector = &selectors[sel];
	if(selector->sockets == NULL)
	{
		selector->sockets_capacity = 8;
		selector->sockets = ALLOC(8 * sizeof(octa));
	}
	if(selector->sockets_length >= selector->sockets_capacity)
	{
		longo old_capacity = selector->sockets_capacity;
		selector->sockets_capacity *= 2;
		selector->sockets = REALLOC(selector->sockets, selector->sockets_capacity * sizeof(octa), old_capacity * sizeof(octa));
	}
	selector->sockets[selector->sockets_length] = s;
	selector->sockets_length++;
}
void network_selector_remove(longo sel, octa s)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	Network_Selector* selector = &selectors[sel];
	int found = INVALID_SELECTOR;
	for(int i = 0; i < selector->sockets_length; i++)
	{		
		if(selector->sockets[i] == s)
		{
			found = i;
			break;
		}
	}
	if(found != INVALID_SELECTOR)
	{
		selector->sockets_length--;
		selector->sockets[found] = selector->sockets[selector->sockets_length];
	}
}
inta network_selector_wait(longo sel, inta timeout, octa* ready_sockets_buf)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return 0;
	octa* ready_sockets = (octa*)ready_sockets_buf[0];
	inta max_sockets = (inta)ready_sockets_buf[1];
	Network_Selector* selector = &selectors[sel];
	struct timeval tv;
	tv.tv_sec = timeout / 1000;
	tv.tv_usec = (timeout % 1000) * 1000;
	fd_set readfds;
	FD_ZERO(&readfds);
	int maxfd = 0;
	for(int i = 0; i < selector->sockets_length; i++)
	{
		if(selector->sockets[i] > maxfd)
			maxfd = selector->sockets[i];
		FD_SET(selector->sockets[i], &readfds);
	}
	int num_ready = select(maxfd+1, &readfds, NULL, NULL, &tv);
	int pos = 0;	
	for(int i = 0; i < selector->sockets_length; i++)
	{
		if(FD_ISSET(selector->sockets[i], &readfds))
		{
			ready_sockets[pos] = selector->sockets[i];
			pos++;
		}
		if(pos >= max_sockets)
			break;
	}
	return num_ready;
}
#endif
#ifdef LINUX_OS
//simple and efficient solution for linux: epoll, although maybe
//	not as thread-safe as network enthousiasts might want.
typedef struct Network_Selector
{
	byte alive;
	int epfd;
	int num_sockets;
} Network_Selector;
Network_Selector selectors[MAX_SELECTORS];
longo num_selectors;
longo network_selector_create()
{
	longo h = 0;
	for(; h < num_selectors; h++)
		if(!selectors[h].alive)
			break;
	if(h >= MAX_SELECTORS)
		return INVALID_SELECTOR;
	if(h >= num_selectors)
		num_selectors = h+1;
	selectors[h].alive = true;
	//16 is really only a dummy value, as it is ignored in modern implementations
	//and it is just a hint for initial allocation
	selectors[h].epfd = epoll_create(16);
	return 0;
}
void network_selector_destroy(longo sel)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	close(selectors[sel].epfd);
	selectors[sel].alive = false;
}
void network_selector_add(longo sel, octa s)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	selectors[sel].num_sockets++;
	struct epoll_event ev;
	memset(&ev, 0, sizeof(struct epoll_event));
	ev.events = EPOLLIN;
	ev.data.fd = s;
	epoll_ctl(selectors[sel].epfd, EPOLL_CTL_ADD, s, &ev);
}
void network_selector_remove(longo sel, octa s)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return;
	struct epoll_event ev;
	epoll_ctl(selectors[sel].epfd, EPOLL_CTL_DEL, s, &ev);
	selectors[sel].num_sockets--;
}
inta network_selector_wait(longo sel, inta timeout, octa* ready_sockets_buf)
{
	if(sel < 0 || sel >= MAX_SELECTORS || !selectors[sel].alive) return 0;
	octa* ready_sockets = (octa*)ready_sockets_buf[0];
	inta max_sockets = (inta)ready_sockets_buf[1];
	int num = selectors[sel].num_sockets;
	int epfd = selectors[sel].epfd;
	if(max_sockets < num)
		num = max_sockets;
	struct epoll_event* events = (struct epoll_event*)temporary_buffer;
	int num_ready = epoll_wait(epfd, events, num, timeout);
	for(int i = 0; i < num_ready; i++)
		ready_sockets[i] = events[i].data.fd;
	return num_ready;
}
#endif
